import warnings
from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import rioxarray as rio
import xarray as xr
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_array_equal,
    assert_equal,
)
from pyvista import MultiBlock, UnstructuredGrid
from shapely.geometry import LineString, Point, Polygon

from geost import config
from geost._warnings import AlignmentWarning, ValidationWarning
from geost.base import BoreholeCollection, LayeredData, PointHeader
from geost.export import geodataclass


class TestCollection:
    @pytest.fixture
    def valid_boreholes(self):
        nr = np.full(10, "B-01")
        x = np.full(10, 139370)
        y = np.full(10, 455540)
        mv = np.full(10, 1.0)
        end = np.full(10, -4.0)
        top = np.array([0, 0.5, 1, 1.5, 2, 3, 4, 5, 6, 7])
        bottom = np.array([0.5, 1, 1.5, 2, 3, 4, 5, 6, 7, 8])
        data_string = np.array(
            ["K", "Kz", "K", "Ks3", "Ks2", "V", "Zk", "Zs", "Z", "Z"]
        )
        data_int = np.arange(0, 10, dtype=np.int64)
        data_float = np.arange(0, 5, 0.5, dtype=np.float64)

        df = pd.DataFrame(
            {
                "nr": nr,
                "x": x,
                "y": y,
                "surface": mv,
                "end": end,
                "top": top,
                "bottom": bottom,
                "data_string": data_string,
                "data_int": data_int,
                "data_float": data_float,
            }
        )

        return LayeredData(df)

    @pytest.fixture
    def invalid_borehole_table(self):
        nr = np.full(10, 1)
        x = np.full(10, 139370)
        y = np.full(10, 455540)
        mv = np.full(10, 1.0)
        end = np.full(10, 4.0)
        top = np.array([0, 0.5, 1, 1.5, 2, 3, 4, 5, 6, 7])
        bottom = np.array([0.5, 1, 1.5, 2, 3, 4, 5, 6, 7, 6.5])
        data_string = np.array(
            ["K", "Kz", "K", "Ks3", "Ks2", "V", "Zk", "Zs", "Z", "Z"]
        )
        data_int = np.arange(0, 10, dtype=np.int64)
        data_float = np.arange(0, 5, 0.5, dtype=np.float64)

        return pd.DataFrame(
            {
                "nr": nr,
                "x": x,
                "y": y,
                "surface": mv,
                "end": end,
                "top": top,
                "bottom": bottom,
                "data_string": data_string,
                "data_int": data_int,
                "data_float": data_float,
            }
        )

    @pytest.fixture
    def misaligned_header(self):
        df = gpd.GeoDataFrame(
            {
                "nr": ["B-02"],
                "x": [100000],
                "y": [400000],
                "surface": [0],
                "end": [-8],
            },
            geometry=[Point(100000, 400000)],
        )

        return PointHeader(df, 28992)

    @pytest.fixture
    def update_raster(self):
        x_coors = np.arange(127000, 128500, 500)
        y_coors = np.arange(503000, 501500, -500)
        data = np.ones((3, 3))
        array = xr.DataArray(data, {"x": x_coors, "y": y_coors})
        return array

    @pytest.mark.unittest
    def test_get_single_object(self, borehole_collection):
        borehole_collection_single_selection = borehole_collection.get("A")
        assert borehole_collection_single_selection.header.gdf.iloc[0, 0] == "A"

    @pytest.mark.unittest
    def test_get_multiple_objects(self, borehole_collection):
        borehole_collection_multi_selection = borehole_collection.get(["A", "B"])
        assert list(borehole_collection_multi_selection.header["nr"].unique()) == [
            "A",
            "B",
        ]

    @pytest.mark.unittest
    def test_add_header_column_to_data(self, borehole_collection):
        borehole_collection.header["test_data"] = [
            i for i in range(len(borehole_collection.header))
        ]
        borehole_collection.add_header_column_to_data("test_data")

        assert_allclose(borehole_collection.get("A").data["test_data"], 0)
        assert_allclose(borehole_collection.get("B").data["test_data"], 1)

    @pytest.mark.unittest
    def test_change_vertical_reference(self, borehole_collection):
        assert borehole_collection.vertical_reference == 5709
        borehole_collection.change_vertical_reference("Ostend height")
        assert borehole_collection.vertical_reference == 5710

    @pytest.mark.unittest
    def test_change_horizontal_reference(self, borehole_collection):
        assert borehole_collection.horizontal_reference == 28992
        borehole_collection.change_horizontal_reference(32631)
        assert borehole_collection.horizontal_reference == 32631
        assert_almost_equal(
            borehole_collection.data["x"][0], borehole_collection.header["x"][0]
        )
        assert_almost_equal(borehole_collection.data["x"][0], 523403.281803, decimal=5)
        assert_almost_equal(
            borehole_collection.data["y"][0], borehole_collection.header["y"][0]
        )
        assert_almost_equal(borehole_collection.data["y"][0], 5313546.187669, decimal=5)

    @pytest.mark.unittest
    def test_change_horizontal_reference_also_data_columns_inclined(
        self, nlog_borehole_collection
    ):
        assert nlog_borehole_collection.horizontal_reference == 28992
        nlog_borehole_collection.change_horizontal_reference(32631)
        assert nlog_borehole_collection.horizontal_reference == 32631
        assert_almost_equal(
            nlog_borehole_collection.header["x"][0], 532641.76, decimal=2
        )
        assert_almost_equal(
            nlog_borehole_collection.header["y"][0], 6.170701e6, decimal=0
        )
        assert_almost_equal(
            nlog_borehole_collection.data["x_bot"][0], 532650.77, decimal=2
        )
        assert_almost_equal(
            nlog_borehole_collection.data["y_bot"][0], 6.170700e6, decimal=0
        )

    @pytest.mark.unittest
    def test_select_within_bbox(self, borehole_collection):
        borehole_collection_selected = borehole_collection.select_within_bbox(
            1.5, 1.5, 3.5, 5
        )
        # The selection results boreholes 'A' and 'D'
        assert all(borehole_collection_selected.header.gdf["nr"].unique() == ["A", "D"])
        assert all(borehole_collection_selected.data.df["nr"].unique() == ["A", "D"])

    @pytest.mark.unittest
    def test_select_with_points(self, borehole_collection):
        selection_points = [Point(1, 1), Point(4, 4), Point(1, 4), Point(4, 1)]
        selection_gdf = gpd.GeoDataFrame(
            {"id": [0, 1, 2, 3]}, geometry=selection_points
        )
        collection_sel = borehole_collection.select_with_points(selection_gdf, 1.1)
        collection_sel_inverted = borehole_collection.select_with_points(
            selection_gdf, 1.1, invert=True
        )
        assert collection_sel.n_points == 3
        assert collection_sel_inverted.n_points == 2

    @pytest.mark.unittest
    def test_select_with_lines(self, borehole_collection):
        selection_lines = [LineString([[1, 1], [5, 5]]), LineString([[1, 5], [5, 1]])]
        selection_gdf = gpd.GeoDataFrame({"id": [0, 1]}, geometry=selection_lines)
        collection_sel = borehole_collection.select_with_lines(selection_gdf, 0.6)
        collection_sel_inverted = borehole_collection.select_with_lines(
            selection_gdf, 0.6, invert=True
        )
        assert collection_sel.n_points == 2
        assert collection_sel_inverted.n_points == 3

    @pytest.mark.unittest
    def test_select_within_polygons(self, borehole_collection):
        selection_polygon = [Polygon(((2, 1), (5, 4), (4, 5), (1, 3)))]
        selection_gdf = gpd.GeoDataFrame({"id": [0]}, geometry=selection_polygon)

        # Polygon selection without buffer
        collection_sel = borehole_collection.select_within_polygons(selection_gdf)
        collection_sel_inverted = borehole_collection.select_within_polygons(
            selection_gdf, invert=True
        )
        assert collection_sel.n_points == 1
        assert collection_sel_inverted.n_points == 4

        # Polygon selection with a buffer
        collection_sel_buff = borehole_collection.select_within_polygons(
            selection_gdf, buffer=0.7
        )
        collection_sel_inverted_buff = borehole_collection.select_within_polygons(
            selection_gdf, buffer=0.7, invert=True
        )
        assert collection_sel_buff.n_points == 2
        assert collection_sel_inverted_buff.n_points == 3

    @pytest.mark.unittest
    def test_select_by_depth(self, borehole_collection):
        assert borehole_collection.select_by_depth(top_min=0).n_points == 4
        assert borehole_collection.select_by_depth(top_max=0).n_points == 1
        assert borehole_collection.select_by_depth(end_min=-3.5).n_points == 2
        assert (
            borehole_collection.select_by_depth(
                top_min=0, top_max=0.3, end_max=-4
            ).n_points
            == 2
        )

    @pytest.mark.unittest
    def test_select_by_length(self, borehole_collection):
        assert borehole_collection.select_by_length(min_length=4).n_points == 2
        assert borehole_collection.select_by_length(max_length=4).n_points == 3
        assert (
            borehole_collection.select_by_length(min_length=3, max_length=5).n_points
            == 4
        )

    @pytest.mark.unittest
    def test_select_by_values(self, borehole_collection):
        selected = borehole_collection.select_by_values("lith", ["V", "K"], how="or")

        expected_nrs = ["A", "B", "C", "D"]
        selected_nrs = selected.data["nr"].unique()

        assert_array_equal(selected_nrs, expected_nrs)

        selected = borehole_collection.select_by_values("lith", ["V", "K"], how="and")

        expected_nrs = ["B", "D"]
        selected_nrs = selected.data["nr"].unique()

        assert_array_equal(selected_nrs, expected_nrs)

    @pytest.mark.unittest
    def test_slice_depth_interval(self, borehole_collection):
        # Test slicing with respect to depth below the surface.
        upper, lower = 0.6, 2.4
        sliced = borehole_collection.slice_depth_interval(upper, lower)

        layers_per_borehole = sliced.data["nr"].value_counts()
        expected_layer_count = [3, 3, 3, 3, 2]

        assert len(sliced.data) == 14
        assert sliced.n_points == 5
        assert sliced.data["top"].min() == upper
        assert sliced.data["bottom"].max() == lower
        assert_array_equal(layers_per_borehole, expected_layer_count)

        # Test slicing without updating layer boundaries.
        sliced = borehole_collection.slice_depth_interval(
            upper, lower, update_layer_boundaries=False
        )

        expected_tops_of_slice = [0.0, 0.6, 0.0, 0.5, 0.5]
        expected_bottoms_of_slice = [2.5, 2.5, 2.9, 2.5, 2.5]

        tops_of_slice = sliced.data.df.groupby("nr")["top"].min()
        bottoms_of_slice = sliced.data.df.groupby("nr")["bottom"].max()

        assert len(sliced.data) == 14
        assert sliced.n_points == 5
        assert_array_equal(tops_of_slice, expected_tops_of_slice)
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

        # Test slicing with respect to a vertical reference plane.
        nap_upper, nap_lower = -2, -3
        sliced = borehole_collection.slice_depth_interval(
            nap_upper, nap_lower, relative_to_vertical_reference=True
        )

        expected_tops_of_slice = [2.2, 2.3, 2.25, 2.1, 1.9]
        expected_bottoms_of_slice = [3.2, 3.3, 3.25, 3.0, 2.9]

        tops_of_slice = sliced.data.df.groupby("nr")["top"].min()
        bottoms_of_slice = sliced.data.df.groupby("nr")["bottom"].max()

        assert len(sliced.data) == 11
        assert sliced.n_points == 5
        assert_array_equal(tops_of_slice, expected_tops_of_slice)
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

        # Test slices that return empty objects.
        empty_slice = borehole_collection.slice_depth_interval(-2, -1)
        empty_slice_nap = borehole_collection.slice_depth_interval(
            3, 2, relative_to_vertical_reference=True
        )

        assert len(empty_slice.data) == 0
        assert len(empty_slice_nap.data) == 0
        assert empty_slice.n_points == 0
        assert empty_slice_nap.n_points == 0

        # Test slicing using only an upper boundary or lower boundary.
        upper = 4
        sliced = borehole_collection.slice_depth_interval(upper)

        expected_boreholes = ["A", "C"]

        assert len(sliced.data) == 2
        assert sliced.n_points == 2
        assert_array_equal(sliced.data["nr"], expected_boreholes)

        nap_lower = -0.5
        sliced = borehole_collection.slice_depth_interval(
            lower_boundary=nap_lower, relative_to_vertical_reference=True
        )

        bottoms_of_slice = sliced.data.df.groupby("nr")["bottom"].max()
        expected_bottoms_of_slice = [0.7, 0.8, 0.75, 0.6, 0.4]

        assert len(sliced.data) == 7
        assert sliced.n_points == 5
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

    @pytest.mark.unittest
    def test_slice_by_values(self, borehole_collection):
        sliced = borehole_collection.slice_by_values("lith", "Z")

        expected_boreholes_with_sand = ["A", "C", "D", "E"]
        expected_length = 10

        assert_array_equal(sliced.data["nr"].unique(), expected_boreholes_with_sand)
        assert np.all(sliced.data["lith"] == "Z")
        assert len(sliced.data) == expected_length
        assert sliced.n_points == 4

        sliced = borehole_collection.slice_by_values("lith", "Z", invert=True)

        expected_boreholes_without_sand = ["A", "B", "C", "D"]
        expected_length = 15

        assert_array_equal(sliced.data["nr"].unique(), expected_boreholes_without_sand)
        assert ~np.any(sliced.data["lith"] == "Z")
        assert len(sliced.data) == expected_length
        assert sliced.n_points == 4

    @pytest.mark.unittest
    def test_select_by_condition(self, borehole_collection):
        # same selection with this method is used in test_data_objects
        selected = borehole_collection.select_by_condition(
            borehole_collection.data["lith"] == "V"
        )
        assert isinstance(selected, BoreholeCollection)
        assert selected.n_points == 2

    @pytest.mark.unittest
    def test_validation_pass(self, valid_boreholes):
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            valid_boreholes.to_collection()

    @pytest.mark.unittest
    def test_validation_fail(self, invalid_borehole_table):
        config.validation.VERBOSE = True
        config.validation.DROP_INVALID = False
        config.validation.FLAG_INVALID = False
        with pytest.warns(ValidationWarning) as record:
            LayeredData(invalid_borehole_table).to_collection()

        assert len(record) == 2

        data_result, header_result = record
        assert "Validation failed for schema 'Layer data non-inclined'" in str(
            data_result.message
        )
        assert (
            "DataFrameSchema 'Layer data non-inclined' failed element-wise validator number 0:"
            in str(data_result.message)
        )
        assert "Validation failed for schema 'Point header'" in str(
            header_result.message
        )
        assert (
            "DataFrameSchema 'Point header' failed element-wise validator number 0:"
            in str(header_result.message)
        )

    @pytest.mark.unittest
    def test_validation_fail_drop_invalid(self, invalid_borehole_table):
        config.validation.VERBOSE = True
        config.validation.DROP_INVALID = True
        config.validation.FLAG_INVALID = False
        with pytest.warns(ValidationWarning) as record:
            LayeredData(invalid_borehole_table).to_collection()

        assert len(record) == 1

        assert (
            "Validation dropped 10 row(s) for schema 'Layer data non-inclined'"
            in str(record[0].message)
        )

    @pytest.mark.unittest
    def test_validation_fail_flag_invalid(self, invalid_borehole_table):
        config.validation.VERBOSE = True
        config.validation.DROP_INVALID = False
        config.validation.FLAG_INVALID = True
        with pytest.warns(ValidationWarning) as record:
            collection_flagged = LayeredData(invalid_borehole_table).to_collection()

        assert len(record) == 2
        assert not collection_flagged.header.gdf.loc[0, "is_valid"]
        assert not collection_flagged.data.df.loc[9, "is_valid"]

        data_result, header_result = record
        assert "Validation failed for schema 'Layer data non-inclined'" in str(
            data_result.message
        )
        assert (
            "DataFrameSchema 'Layer data non-inclined' failed element-wise validator number 0:"
            in str(data_result.message)
        )
        assert "Validation failed for schema 'Point header'" in str(
            header_result.message
        )
        assert (
            "DataFrameSchema 'Point header' failed element-wise validator number 0:"
            in str(header_result.message)
        )

    @pytest.mark.unittest
    def test_header_mismatch_auto_align(self, valid_boreholes, misaligned_header):
        config.validation.AUTO_ALIGN = True
        with pytest.warns(AlignmentWarning) as record:
            BoreholeCollection(misaligned_header, valid_boreholes)
        assert len(record) == 2
        assert "Header covers more/other objects than present in the data table" in str(
            record[0].message
        )
        assert "Header does not cover all unique objects in data" in str(
            record[1].message
        )

    @pytest.mark.unittest
    def test_header_mismatch_no_auto_align(self, valid_boreholes, misaligned_header):
        config.validation.AUTO_ALIGN = False
        with pytest.warns(AlignmentWarning) as record:
            BoreholeCollection(misaligned_header, valid_boreholes)
        assert len(record) == 2
        assert "Header covers more/other objects than present in the data table" in str(
            record[0].message
        )
        assert "Header does not cover all unique objects in data" in str(
            record[1].message
        )

    @pytest.mark.unittest
    def test_get_area_labels(self, borehole_collection):
        label_polygon = [Polygon(((2, 1), (5, 4), (4, 5), (1, 4)))]
        label_gdf = gpd.GeoDataFrame({"id": [1]}, geometry=label_polygon, crs=28992)
        # Return variant
        output = borehole_collection.get_area_labels(label_gdf, "id")
        assert_almost_equal(output["id"].sum(), 2)
        # In-place variant
        borehole_collection.get_area_labels(label_gdf, "id", include_in_header=True)
        assert_almost_equal(borehole_collection.header["id"].sum(), 2)

    @pytest.mark.unittest
    def test_get_area_labels_multiple(self, borehole_collection):
        label_polygon = [Polygon(((2, 1), (5, 4), (4, 5), (1, 4)))]
        label_gdf = gpd.GeoDataFrame(
            {"id": [1], "col2": ["string_data"]}, geometry=label_polygon, crs=28992
        )
        # Return variant
        output = borehole_collection.get_area_labels(label_gdf, ["id", "col2"])
        assert_almost_equal(output["id"].sum(), 2)
        assert_equal(output["col2"].value_counts()["string_data"], 2)
        # In-place variant
        borehole_collection.get_area_labels(
            label_gdf, ("id", "col2"), include_in_header=True
        )
        assert_almost_equal(borehole_collection.header["id"].sum(), 2)
        assert_equal(
            borehole_collection.header["col2"].value_counts()["string_data"], 2
        )

    @pytest.mark.unittest
    def test_to_datafusiontools(self, borehole_collection):
        # More detailed tests are in TestLayeredData in test_data_objects.py
        dft = borehole_collection.to_datafusiontools("lith")
        assert np.all([isinstance(d, geodataclass.Data) for d in dft])

        outfile = Path("dft.pkl")
        borehole_collection.to_datafusiontools("lith", outfile)
        assert outfile.is_file()
        outfile.unlink()

    @pytest.mark.unittest
    def test_to_qgis3d(self, borehole_collection):
        outfile = Path("temp.gpkg")
        borehole_collection.to_qgis3d(outfile)
        assert outfile.is_file()
        outfile.unlink()

    @pytest.mark.unittest
    def test_to_geoparquet(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.geoparquet"
        borehole_collection.to_geoparquet(outfile)
        assert outfile.is_file()

    @pytest.mark.unittest
    def test_to_shape(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.shp"
        borehole_collection.to_shape(outfile)
        assert outfile.is_file()

    @pytest.mark.unittest
    def test_to_geopackage(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.gpkg"
        borehole_collection.to_geopackage(outfile)
        assert outfile.is_file()
        layers = gpd.list_layers(outfile)
        assert_array_equal(layers["name"], ["header", "data"])
        assert_array_equal(layers["geometry_type"], ["Point", None])

    @pytest.mark.unittest
    def test_to_parquet(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.gpkg"
        borehole_collection.to_parquet(outfile)
        assert outfile.is_file()

    @pytest.mark.unittest
    def test_to_csv(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.gpkg"
        borehole_collection.to_csv(outfile)
        assert outfile.is_file()

    @pytest.mark.unittest
    def test_to_pickle(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.pkl"
        borehole_collection.to_pickle(outfile)
        assert outfile.is_file()

    # @pytest.mark.integrationtest
    # def test_surface_level_update(self, borehole_collection, update_raster):
    #     borehole_collection.update_surface_level_from_raster(update_raster, how="replace")
    #     print("stop")


class TestBoreholeCollection:
    @pytest.mark.unittest
    def test_get_cumulative_thickness_multiple(self, borehole_collection):
        borehole_collection.get_cumulative_thickness(
            "lith", ["Z", "K"], include_in_header=True
        )
        expected_clay_thickness = [2.0, 2.0, 2.9, 1.1, 0.0]
        expected_sand_thickness = [2.2, 0.0, 2.6, 0.5, 3.0]

        assert_almost_equal(
            borehole_collection.header["K_thickness"], expected_clay_thickness
        )
        assert_almost_equal(
            borehole_collection.header["Z_thickness"], expected_sand_thickness
        )

        # Single query
        borehole_collection.get_cumulative_thickness(
            "lith", "Z", include_in_header=True
        )
        assert_almost_equal(
            borehole_collection.header["Z_thickness"], expected_sand_thickness
        )

    @pytest.mark.unittest
    def test_get_cumulative_thickness_single(self, borehole_collection):
        expected_sand_thickness = [2.2, 0.0, 2.6, 0.5, 3.0]

        # Single query
        borehole_collection.get_cumulative_thickness(
            "lith", "Z", include_in_header=True
        )
        assert_almost_equal(
            borehole_collection.header["Z_thickness"], expected_sand_thickness
        )

    @pytest.mark.unittest
    def test_get_layer_top(self, borehole_collection):
        borehole_collection.get_layer_top("lith", ["Z", "K"], include_in_header=True)

        expected_sand_top = [1.5, np.nan, 2.9, 2.5, 0.0]
        expected_clay_top = [0.0, 0.0, 0.0, 0.0, np.nan]

        assert_almost_equal(borehole_collection.header["K_top"], expected_clay_top)
        assert_almost_equal(borehole_collection.header["Z_top"], expected_sand_top)

    @pytest.mark.unittest
    def test_to_kingdom(self, borehole_collection):
        outfile = Path("temp_kingdom.csv")
        tdfile = Path(outfile.parent, f"{outfile.stem}_TDCHART{outfile.suffix}")
        borehole_collection.to_kingdom(outfile)
        assert outfile.is_file()
        assert tdfile.is_file()
        outfile.unlink()
        tdfile.unlink()


class TestCptCollection:
    @pytest.mark.unittest
    def test_slice_depth_interval(self, cpt_collection):
        upper, lower = 0.6, 4.4
        sliced = cpt_collection.slice_depth_interval(upper, lower)

        assert len(sliced.data) == 8
        assert sliced.n_points == 2
        assert sliced.data["depth"].min() >= upper
        assert sliced.data["depth"].max() <= lower

        upper, lower = 1.9, 0.9  # Elevations in NAP
        sliced = cpt_collection.slice_depth_interval(
            upper, lower, relative_to_vertical_reference=True
        )
        assert sliced.n_points == 1
        assert len(sliced.data) == 1



from pathlib import Path

import numpy as np
import pandas as pd
from lxml import etree

from geost.io.parsers.parser_utils import safe_coerce


class BorisXML:
    """
    Class with properties and reading/parsing methods for BORIS XML exports. BORIS is
    software developed by TNO to describe boreholes in the lab. The exported XML-format
    bears no resemblance to DINO or BRO XML standards.

    The reader is naive, which means that all data is taken from the XML as is, without
    running any checks on the parsed data.

    Note that currently only essential items to construct a BoreholeCollection are parsed.
    """

    attr_type_to_dtype = dict(
        code=str,
        median=float,
        percentage=float,
    )

    def __init__(self, xml: str | Path | etree._Element):
        if isinstance(xml, (str, Path)):
            self.root = etree.parse(xml).getroot()
        elif isinstance(xml, etree._Element):
            self.root = xml
        else:
            raise TypeError(
                "Input datatype for reader not allowed. Must be an lxml Element "
                "tree or path to an xml file."
            )

        self.layer_dataframes = list()

        for point_survey in self.root.iterchildren():
            self.parse_pointsurvey(point_survey)
        self.layer_dataframe = pd.concat(
            self.layer_dataframes, ignore_index=True
        ).convert_dtypes(convert_integer=False)

    def parse_pointsurvey(self, point_survey: etree._Element) -> None:
        # Get header and layer data as dictionaries
        header_data = self.parse_headerdata(point_survey)
        layer_dataframe = self.parse_layerdata(
            point_survey.findall("borehole/lithoDescr/lithoInterval")
        )
        layer_dataframe.insert(
            1,
            "top",
            [0] + list(layer_dataframe["bottom"].iloc[:-1].values),
        )
        layer_dataframe["end"] += header_data["surface"]

        # Repeat header data for every described borehole interval
        for key, value in header_data.items():
            layer_dataframe.insert(0, key, np.full(len(layer_dataframe), value))

        # Append data of this borehole to
        self.layer_dataframes.append(layer_dataframe)

    def parse_headerdata(self, point_survey: etree._Element) -> dict:
        # Find elements that contain header data
        identification_element = point_survey.find("identification")
        location_element = point_survey.find("surveyLocation")
        elevation_element = point_survey.find("surfaceElevation/elevation")

        # Get positional reference system attributes
        # self.horizontal_reference = location_element.find("coordinates").get(
        #     "coordSystem"
        # )
        # self.vertical_reference = elevation_element.get("levelReference") or "NAP"

        # Get header data. Note that the 'end' column is later retrieved and added to
        # the header based on the last layer interval.
        nr = safe_coerce(identification_element.get("id"), str)
        x = safe_coerce(location_element.find("coordinates/coordinateX").text, float)
        y = safe_coerce(location_element.find("coordinates/coordinateY").text, float)
        surface = safe_coerce(elevation_element.get("levelValue"), float)

        # Create header data dict
        header_data = dict(surface=surface, y=y, x=x, nr=nr)
        return header_data

    def parse_layerdata(self, borehole_element: etree._Element) -> pd.DataFrame:
        *_, last = borehole_element
        end = safe_coerce(last.get("baseDepth"), float) * -1

        num_of_layers = len(borehole_element)
        layer_data = list()
        unique_tags = ["bottom"]
        for interval in borehole_element:
            interval_dict = dict()
            interval_dict["bottom"] = float(interval.get("baseDepth"))
            for data_element in interval.getchildren():
                if data_element.tag not in unique_tags:
                    unique_tags += [data_element.tag]
                if data_element.attrib:
                    attr_type, attr_value = (
                        data_element.attrib.keys()[0],
                        data_element.attrib.values()[0],
                    )
                    interval_dict[data_element.tag] = safe_coerce(
                        attr_value, self.attr_type_to_dtype[attr_type]
                    )
                else:
                    interval_dict[data_element.tag] = safe_coerce(
                        data_element.text, str
                    )
            layer_data.append(interval_dict)

        # Construct final dataframe
        layer_df = pd.DataFrame(columns=unique_tags, index=range(num_of_layers))
        for i, layer in enumerate(layer_data):
            layer_df.update(pd.DataFrame(layer, index=[i]))
        layer_df.insert(0, "end", np.full(num_of_layers, end))
        return layer_df



from typing import Literal

import pytest
from lxml.etree import _Element

from geost.bro import BroApi


class TestBroApi:
    @pytest.mark.unittest
    def test_response(self):
        api = BroApi()
        for key in api.apis:
            # Exclude BRO Geological boreholes until added to BRO database
            # (They currently return a 503 status code)
            if key != "BHR-G":
                assert (
                    api.session.get(api.server_url + api.apis[key]).status_code == 200
                )

    @pytest.mark.parametrize(
        "objects, object_type",
        (
            (["CPT000000038771", "CPT000000000787", "CPT000000125133"], "CPT"),
            (["BHR000000123271", "BHR000000319277", "BHR000000051413"], "BHR-P"),
            (["BHR000000347437", "BHR000000347438", "BHR000000347434"], "BHR-GT"),
            (["BHR000000396339", "BHR000000396340", "BHR000000396341"], "BHR-G"),
            (["SFR000000002177", "SFR000000002172", "SFR000000002655"], "SFR"),
        ),
        ids=["CPT", "BHR-P", "BHR-GT", "BHR-G", "SFR"],
    )
    def test_get_valid_objects(self, objects, object_type):
        api = BroApi()
        objects = api.get_objects(objects, object_type=object_type)
        assert isinstance(objects, list)
        assert len(objects) == 3
        for obj in objects:
            assert isinstance(obj, bytes)

    @pytest.mark.parametrize(
        "bro_id, object_type, expected_exception",
        (
            ("NonExistingBroId", "CPT", "Unable to request NonExistingBroId"),
            ("NonExistingBroId", "BHR-GT", "Unable to request NonExistingBroId"),
        ),
    )
    def test_get_invalid_objects(
        self,
        bro_id: Literal["NonExistingBroId"],
        object_type: Literal["CPT"] | Literal["BHR-GT"],
        expected_exception: Literal["Unable to request NonExistingBroId"],
    ):
        api = BroApi()
        with pytest.raises(Warning) as excinfo:
            api.get_objects(bro_id, object_type=object_type)
            assert expected_exception in str(excinfo.value)

    @pytest.mark.unittest
    def test_search_and_get_cpts_in_bbox(self):
        # Note: if test fails, check if BRO database was updated first
        api = BroApi()
        api.search_objects_in_bbox(112400, 442750, 112500, 442850, object_type="CPT")
        minimum_present_objects = ["CPT000000000787", "CPT000000029403"]
        present_objects = [
            obj for obj in api.object_list if obj in minimum_present_objects
        ]
        assert len(present_objects) == len(minimum_present_objects)

    @pytest.mark.unittest
    def test_search_and_get_bhrps_in_bbox(self):
        # Note: if test fails, check if BRO database was updated first
        api = BroApi()

        api.search_objects_in_bbox(141500, 455100, 141700, 455300, object_type="BHR-P")
        minimum_present_objects = [
            "BHR000000085497",
            "BHR000000247842",
            "BHR000000120513",
            "BHR000000206176",
        ]
        present_objects = [
            obj for obj in api.object_list if obj in minimum_present_objects
        ]
        assert len(present_objects) == len(minimum_present_objects)

    @pytest.mark.unittest
    def test_search_and_get_bhrgts_in_bbox(self):
        # Note: if test fails, check if BRO database was updated first
        api = BroApi()
        # Find CPTs
        api.search_objects_in_bbox(141300, 452700, 142300, 453500, object_type="BHR-GT")
        minimum_present_objects = [
            "BHR000000353592",
            "BHR000000353583",
            "BHR000000353598",
            "BHR000000353600",
        ]
        present_objects = [
            obj for obj in api.object_list if obj in minimum_present_objects
        ]
        assert len(present_objects) == len(minimum_present_objects)

    @pytest.mark.unittest
    def test_get_too_large_volume(self, capsys: pytest.CaptureFixture[str]):
        api = BroApi()
        _ = api.search_objects_in_bbox(82000, 425000, 90000, 432000)
        captured = capsys.readouterr()
        assert "More than 2000 object requests in API call" in captured.out



import numpy as np
import pandas as pd


def find_top_sand(
    lith: np.ndarray,
    top: np.ndarray,
    bottom: np.ndarray,
    min_sand_frac: float,
    min_sand_thickness: float,
) -> float:
    """
    Find the top of sand depth in a borehole described in NEN5104 format. The top of sand
    is defined by the first layer of a specified thickness that contains a minimum
    percentage of sand. By default: when the first layer of sand is detected, the next 1
    meter is scanned. Within this meter, if more than 50% of the lenght has a main
    lithology of sand, the initially detected layer of sand is regarded as the top
    of sand. If not, continue downward until the next layer of sand is detected and
    repeat.

    Parameters
    ----------
    lith : ndarray
        Numpy array containing the lithology of the borehole.
    top : ndarray
        Numpy array containing the top depth of the layers of the borehole.
    bottom : ndarray
        Numpy array containing the bottom depth of the layers of the borehole.
    min_sand_frac : float
        Minimum percentage required to be sand.
    min_sand_thickness : float
        Minimum thickness of the sand to search for.

    Returns
    -------
    top_sand : float
        Top depth of the sand layer that meets the requirements.

    """
    is_sand = ("Z" == lith) + ("G" == lith)

    found_sand = False
    if np.any(is_sand):
        idx_sand = np.flatnonzero(is_sand)
        for idx in idx_sand:
            top_sand = top[idx]
            search_depth = top_sand + min_sand_thickness

            search_mask = (top >= top_sand) & (top < search_depth)

            tmp_top = top[search_mask].copy()
            tmp_bottom = bottom[search_mask].copy()

            if tmp_bottom[-1] > search_depth:
                tmp_bottom[-1] = search_depth

            length = tmp_bottom - tmp_top

            sand_frac = length[is_sand[search_mask]].sum() / min_sand_thickness

            if sand_frac >= min_sand_frac:
                found_sand = True
                break

    if not found_sand:
        top_sand = np.nan

    return top_sand


def top_of_sand(
    boreholes: pd.DataFrame,
    ids: str = "nr",
    min_sand_frac: float = 0.5,
    min_sand_thickness: int | float = 1,
):
    """
    Find the top of sand depth in a borehole described in NEN5104 format. The top of sand
    is defined by the first layer of a specified thickness that contains a minimum fraction
    of sand.

    Parameters
    ----------
    boreholes : pd.DataFrame
        Boreholes in NEN5104 format with "lith", "top" and "bottom" columns.
    ids : str, optional
        Column specifying the borehole IDs. The default is "nr".
    min_sand_frac : float, optional
        Minimum percentage of sand in the sand layer. The default is 0.5 (=50%).
    min_sand_thickness : int | float, optional
        Minimum thickness of the sand layer to find the top of. The default is 1.

    Returns
    ------
    pd.DataFrame
        DataFrame containing the borehole IDs and the top of sand depths.

    """
    groupby = boreholes.groupby(ids)

    result = []
    for nr, df in groupby:
        lith = df["lith"].values
        top = df["top"].values
        bottom = df["bottom"].values

        top_sand = find_top_sand(lith, top, bottom, min_sand_frac, min_sand_thickness)

        result.append((nr, top_sand))

    return pd.DataFrame(result, columns=["nr", "top"])


def cumulative_thickness(data, top: str = "top", bottom: str = "bottom"):
    return np.abs(np.sum(data[top] - data[bottom]))


def layer_top(data, column: str, value: str):  # TODO
    for nr, obj in data.groupby("nr"):
        try:
            layer_top = obj[obj[column] == value].iloc[0]["top"]
        except IndexError:
            layer_top = np.nan
        yield (nr, layer_top)



from geost.base import LineHeader, PointHeader


class HeaderFactory:
    def __init__(self):
        self._types = {}

    def register_header(self, type_, header):
        self._types[type_] = header

    @staticmethod
    def get_geometry_type(gdf):
        geometry_type = gdf.geom_type.unique()
        if len(geometry_type) == 0:
            return None
        elif len(geometry_type) > 1:
            return "Multiple geometries"
        else:
            return geometry_type[0]

    def create_from(self, gdf, **kwargs):
        geometry_type = self.get_geometry_type(gdf)
        header = self._types.get(geometry_type)
        if not header:
            raise ValueError(f"No Header type available for {geometry_type}")
        return header(gdf, **kwargs)


header_factory = HeaderFactory()
header_factory.register_header("Point", PointHeader)
header_factory.register_header("Line", LineHeader)



import warnings
from pathlib import Path

import pandas as pd
import pytest
from numpy.testing import assert_array_equal
from shapely import geometry as gmt

import geost
from geost.base import (
    BoreholeCollection,
    CptCollection,
    DiscreteData,
    LayeredData,
    PointHeader,
)
from geost.io.read import (
    MANDATORY_LAYERED_DATA_COLUMNS,
    _check_mandatory_column_presence,
    adjust_z_coordinates,
)


class InvalidCollection:
    """
    Invalid Collection type to raise a ValueError in `read_collection_geopackage`.
    """

    pass


@pytest.fixture
def llg_header_table(tmp_path):
    outfile = tmp_path / r"llg_header.parquet"
    header = pd.DataFrame(
        {
            "BOORP": [1901, 1902],
            "XCO": [1, 2],
            "YCO": [1, 2],
            "MV_HoogteNAP": [2.1, 1.9],
            "BoringEinddiepteNAP": [1.7, 1.5],
        }
    )
    header.to_parquet(outfile)
    return outfile


@pytest.fixture
def llg_data_table(tmp_path):
    outfile = tmp_path / r"llg_data.parquet"
    data = pd.DataFrame(
        {
            "BOORP": [1901, 1901, 1902, 1902],
            "BEGIN_DIEPTE": [0, 30, 0, 20],
            "EIND_DIEPTE": [30, 40, 20, 40],
            "TEXTUUR": ["ZK", "L", "Z", "K"],
        }
    )
    data.to_parquet(outfile)
    return outfile


@pytest.fixture
def llg_data_table_duplicate_column(tmp_path):
    outfile = tmp_path / r"llg_data_dupl.parquet"
    data = pd.DataFrame(
        {
            "nr": [1901, 1901, 1902, 1902],
            "top": [0, 30, 0, 20],
            "bottom": [30, 40, 20, 40],
            "TEXTUUR": ["ZK", "L", "Z", "K"],
        }
    )
    data.to_parquet(outfile)
    return outfile


@pytest.fixture
def collection_pickle(borehole_collection, tmp_path):
    outfile = tmp_path / r"collection.pkl"
    borehole_collection.to_pickle(outfile)
    return outfile


@pytest.fixture
def table_wrong_columns():
    return pd.DataFrame(
        {
            "nr": ["a", "b"],
            "x": [1, 2],
            "y": [1, 2],
            "maaiveld": [1, 2],
            "end": [1, 1],
            "top": [1, 1],
            "bottom": [2, 2],
        }
    )


@pytest.mark.unittest
def test_nlog_reader_from_parquet(testdatadir):
    nlog_cores = geost.read_nlog_cores(
        testdatadir / r"test_nlog_stratstelsel_20230807.parquet"
    )
    desired_df = pd.DataFrame(
        {
            "nr": ["BA050018", "BA080036", "BA110040"],
            "x": [37395, 44175, 39309],
            "y": [857077, 840614, 833198],
            "surface": [36.7, 39.54, 33.51],
            "end": [-3921.75, -3262.69, -3865.89],
        }
    )
    assert_array_equal(
        nlog_cores.header[["nr", "x", "y", "surface", "end"]], desired_df
    )
    assert nlog_cores.data.has_inclined


@pytest.mark.unittest
def test_read_borehole_table(testdatadir):
    file_pq = testdatadir / r"test_borehole_table.parquet"
    file_csv = testdatadir / r"test_borehole_table.csv"
    cores_pq = geost.read_borehole_table(file_pq)
    cores_csv = geost.read_borehole_table(file_csv)
    assert isinstance(cores_pq, BoreholeCollection)
    assert isinstance(cores_csv, BoreholeCollection)

    cores_pq = geost.read_borehole_table(file_pq, as_collection=False)
    cores_csv = geost.read_borehole_table(file_csv, as_collection=False)
    assert isinstance(cores_pq, LayeredData)
    assert isinstance(cores_csv, LayeredData)


@pytest.mark.unittest
def test_read_inclined_borehole_table(testdatadir):
    file_pq = testdatadir / r"test_inclined_borehole_table.parquet"
    cores_pq = geost.read_borehole_table(file_pq, has_inclined=True)
    assert isinstance(cores_pq, BoreholeCollection)

    cores_pq = geost.read_borehole_table(
        file_pq, has_inclined=True, as_collection=False
    )
    assert isinstance(cores_pq, LayeredData)


@pytest.mark.unittest
def test_check_mandatory_columns(table_wrong_columns):
    column_mapper = {"maaiveld": "surface"}
    table_wrong_columns = _check_mandatory_column_presence(
        table_wrong_columns, MANDATORY_LAYERED_DATA_COLUMNS, column_mapper
    )
    assert_array_equal(table_wrong_columns.columns, MANDATORY_LAYERED_DATA_COLUMNS)


@pytest.mark.unittest
def test_check_mandatory_columns_with_user_input(table_wrong_columns, monkeypatch):
    monkeypatch.setattr("builtins.input", lambda _: "maaiveld")
    table_wrong_columns = _check_mandatory_column_presence(
        table_wrong_columns, MANDATORY_LAYERED_DATA_COLUMNS
    )
    assert_array_equal(table_wrong_columns.columns, MANDATORY_LAYERED_DATA_COLUMNS)


@pytest.mark.unittest
def test_adjust_z_coordinates(testdatadir):
    file = testdatadir / r"test_borehole_table.parquet"
    cores = geost.read_borehole_table(file, as_collection=False)
    cores_df = cores.df.copy()

    # test situation where top and bottoms are already positive downward
    cores_df_adjusted = adjust_z_coordinates(cores_df.copy())
    assert cores_df_adjusted["top"].iloc[0] < cores_df_adjusted["top"].iloc[1]
    assert cores_df_adjusted["bottom"].iloc[0] < cores_df_adjusted["bottom"].iloc[1]

    # Test situation where top and bottoms are given as negative downward
    cores_df["top"] *= -1
    cores_df["bottom"] *= -1

    cores_df_adjusted = adjust_z_coordinates(cores_df.copy())
    assert cores_df_adjusted["top"].iloc[0] < cores_df_adjusted["top"].iloc[1]
    assert cores_df_adjusted["bottom"].iloc[0] < cores_df_adjusted["bottom"].iloc[1]


@pytest.mark.unittest
def test_read_boris_xml(testdatadir):
    boris_collection = geost.read_xml_boris(testdatadir / r"xml/test_boris_xml.xml")
    assert isinstance(boris_collection, BoreholeCollection)
    assert boris_collection.n_points == 16
    assert len(boris_collection.data.df) == 236


@pytest.mark.unittest
def test_read_uullg_table(
    llg_header_table, llg_data_table, llg_data_table_duplicate_column
):
    llg = geost.read_uullg_tables(llg_header_table, llg_data_table)
    assert isinstance(llg, BoreholeCollection)
    expected_header_cols = ["nr", "x", "y", "surface", "end"]
    header_columns = llg.header.gdf.columns
    assert all([c in header_columns for c in expected_header_cols])
    expected_data_cols = ["nr", "x", "y", "surface", "end"]
    data_columns = llg.data.df.columns
    assert all([c in data_columns for c in expected_data_cols])

    llg = geost.read_uullg_tables(llg_header_table, llg_data_table_duplicate_column)
    assert isinstance(llg, BoreholeCollection)
    expected_header_cols = ["nr", "x", "y", "surface", "end"]
    header_columns = llg.header.gdf.columns
    assert all([c in header_columns for c in expected_header_cols])
    expected_data_cols = ["nr", "x", "y", "surface", "end"]
    data_columns = llg.data.df.columns
    assert all([c in data_columns for c in expected_data_cols])


@pytest.mark.unittest
def test_read_gef_cpts(testdatadir):
    files = sorted(Path(testdatadir / "gef").glob("*.gef"))
    cpts = geost.read_gef_cpts(files)
    assert isinstance(cpts, CptCollection)

    expected_cpts_present = [
        "DKMP_D03",
        "AZZ158",
        "CPT000000038871",
        "CPT000000157983",
        "YANGTZEHAVEN CPT 10",
    ]
    assert_array_equal(cpts.header["nr"], expected_cpts_present)


@pytest.mark.unittest
def test_read_cpt_table(testdatadir, monkeypatch):
    monkeypatch.setattr("builtins.input", lambda _: "mv")
    cpts = geost.read_cpt_table(testdatadir / r"test_cpts.parquet")
    assert isinstance(cpts, CptCollection)
    assert cpts.horizontal_reference == 28992
    assert cpts.vertical_reference == 5709

    cpts = geost.read_cpt_table(testdatadir / r"test_cpts.parquet", as_collection=False)
    assert isinstance(cpts, DiscreteData)


@pytest.mark.unittest
def test_read_pickle(collection_pickle):
    collection = geost.read_pickle(collection_pickle)
    assert isinstance(collection, BoreholeCollection)


class TestReadCollectionGeopackage:
    @pytest.fixture
    def borehole_gpkg(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"borehole.gpkg"
        borehole_collection.to_geopackage(outfile)
        return outfile

    @pytest.fixture
    def cpt_gpkg(self, cpt_collection, tmp_path):
        outfile = tmp_path / r"cpt.gpkg"
        cpt_collection.to_geopackage(outfile)
        return outfile

    @pytest.mark.unittest
    def test_read_collection_geopackage_boreholes(self, borehole_gpkg):
        collection = geost.read_collection_geopackage(borehole_gpkg, BoreholeCollection)
        assert isinstance(collection, BoreholeCollection)
        assert isinstance(collection.header, PointHeader)
        assert isinstance(collection.data, LayeredData)

    @pytest.mark.unittest
    def test_read_collection_geopackage_cpts(self, cpt_gpkg):
        collection = geost.read_collection_geopackage(cpt_gpkg, CptCollection)
        assert isinstance(collection, CptCollection)
        assert isinstance(collection.header, PointHeader)
        assert isinstance(collection.data, DiscreteData)

    @pytest.mark.unittest
    def test_read_collection_geopackage_invalid(self, borehole_gpkg):
        with pytest.raises(ValueError):
            geost.read_collection_geopackage(borehole_gpkg, InvalidCollection)


@pytest.mark.parametrize(
    "object_type, options, expected_bro_ids",
    [
        (
            "BHR-GT",
            {"bro_ids": ["BHR000000339682", "BHR000000339733"]},
            ["BHR000000339682", "BHR000000339733"],
        ),
        (
            "BHR-GT",
            {"bbox": (132780.327, 448030.0, 132782.327, 448032.1)},
            ["BHR000000336600"],
        ),
        (
            "BHR-GT",
            {"geometry": gmt.box(132780.327, 448030.0, 132782.327, 448032.1)},
            ["BHR000000336600"],
        ),
        (
            "BHR-GT",
            {"geometry": gmt.Point(132781.327, 448031.1), "buffer": 2},
            ["BHR000000336600"],
        ),
        (
            "BHR-P",
            {"bro_ids": "BHR000000108193"},
            ["BHR000000108193"],
        ),
        (
            "BHR-P",
            {"bbox": (129490, 452254, 129492, 452256)},
            ["BHR000000108193"],
        ),
        (
            "BHR-P",
            {"geometry": gmt.box(129490, 452254, 129492, 452256)},
            ["BHR000000108193"],
        ),
        (
            "BHR-P",
            {"geometry": gmt.Point(129491, 452255), "buffer": 2},
            ["BHR000000108193"],
        ),
        (
            "BHR-G",
            {"bro_ids": "BHR000000396406"},
            ["BHR000000396406"],
        ),
        (
            "BHR-G",
            {"bbox": (126148, 452161, 126150, 452163)},
            ["BHR000000396406"],
        ),
        (
            "BHR-G",
            {"geometry": gmt.box(126148, 452161, 126150, 452163)},
            ["BHR000000396406"],
        ),
        (
            "BHR-G",
            {"geometry": gmt.Point(126149, 452162), "buffer": 2},
            ["BHR000000396406"],
        ),
        (
            "CPT",
            {"bro_ids": "CPT000000155283"},
            ["CPT000000155283"],
        ),
        (
            "CPT",
            {"bbox": (132781.52, 448029.34, 132783.52, 448031.34)},
            ["CPT000000155283"],
        ),
        (
            "CPT",
            {"geometry": gmt.box(132781.52, 448029.34, 132783.52, 448031.34)},
            ["CPT000000155283"],
        ),
        (
            "CPT",
            {"geometry": gmt.Point(132782.52, 448030.34), "buffer": 2},
            ["CPT000000155283"],
        ),
        (
            "SFR",
            {"bro_ids": "SFR000000000687"},
            ["SFR000000000687"],
        ),
        (
            "SFR",
            {"bbox": (132249, 451074, 132251, 451076)},
            ["SFR000000000687"],
        ),
        (
            "SFR",
            {"geometry": gmt.box(132249, 451074, 132251, 451076)},
            ["SFR000000000687"],
        ),
        (
            "SFR",
            {"geometry": gmt.Point(132250, 451075), "buffer": 2},
            ["SFR000000000687"],
        ),
    ],
)
@pytest.mark.unittest
def test_bro_api_read(object_type, options, expected_bro_ids):
    collection = geost.bro_api_read(object_type, **options)
    assert isinstance(collection, geost.base.Collection)
    assert len(collection) == len(expected_bro_ids)
    assert_array_equal(collection.header["nr"], expected_bro_ids)


@pytest.mark.parametrize(
    "object_type",
    [
        "BHR-GT",
        "BHR-G",
        "CPT",
        "BHR-P",
        "SFR",
    ],
)
@pytest.mark.unittest
def test_bro_api_read_no_results(object_type):
    collection = geost.bro_api_read(object_type, bbox=[0, 0, 1, 1])
    assert isinstance(collection, geost.base.Collection)
    assert "<EMPTY COLLECTION>" in str(collection)


@pytest.mark.unittest
def test_bro_api_read_invalid_object_type():
    with pytest.raises(
        ValueError, match="Object type 'INVALID' is not supported for reading."
    ):
        geost.bro_api_read("INVALID", bro_ids=["BHR000000339682"])


@pytest.mark.unittest
def test_read_bhrp(testdatadir):
    file = testdatadir / r"xml/bhrp_bro.xml"
    collection = geost.read_bhrp(file)
    assert isinstance(collection, BoreholeCollection)


@pytest.mark.unittest
def test_read_bhrgt(testdatadir):
    file = testdatadir / r"xml/bhrgt_bro.xml"
    collection = geost.read_bhrgt(file)
    assert isinstance(collection, BoreholeCollection)


@pytest.mark.unittest
def test_read_bhrg(testdatadir):
    file = testdatadir / r"xml/bhrg_bro.xml"
    collection = geost.read_bhrg(file)
    assert isinstance(collection, BoreholeCollection)


@pytest.mark.unittest
def test_read_cpt(testdatadir):
    file = testdatadir / r"xml/cpt_bro.xml"
    collection = geost.read_cpt(file)
    assert isinstance(collection, CptCollection)


@pytest.mark.unittest
def test_read_sfr(testdatadir):
    file = testdatadir / r"xml/sfr_bro.xml"
    collection = geost.read_sfr(file)
    assert isinstance(collection, BoreholeCollection)



import operator
import sqlite3
from pathlib import Path
from typing import Any, Union

import geopandas as gpd
import pandas as pd
from pyogrio.errors import FieldError
from shapely import geometry as gmt

type Geometry = gmt.base.BaseGeometry

COMPARISON_OPERATORS = {
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
}

ARITHMIC_OPERATORS = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
}


def _pandas_read(file: str | Path, **kwargs) -> pd.DataFrame:
    """
    Read parquet file.

    Parameters
    ----------
    file : Path
        Path to file to be read.
    kwargs : optional
        Kwargs to pass to the read function

    Returns
    -------
    pd.DataFrame
        Dataframe with contents of 'file'.

    Raises
    ------
    TypeError
        if 'file' has no valid suffix.

    """
    file = Path(file)
    suffix = file.suffix
    if suffix in [".parquet", ".pq"]:
        return pd.read_parquet(file, **kwargs)
    elif suffix in [".csv", ".txt", ".tsv"]:
        return pd.read_csv(file, **kwargs)
    elif suffix in [".xls", ".xlsx"]:  # pragma: no cover
        return pd.read_excel(
            file, **kwargs
        )  # No cover: Excel file io fails in windows CI pipeline
    else:
        raise TypeError(
            f"Expected parquet file (with .parquet or .pq suffix) but got {suffix} file"
        )


def _geopandas_read(file: str | Path, **kwargs) -> gpd.GeoDataFrame:
    """
    Read a geospatial file into a GeoDataFrame.

    Parameters
    ----------
    file : str | Path
        Path to the geospatial file to be read.
    kwargs : optional
        Additional keyword arguments to pass to the geopandas read function. See relevant
        documentation for details.

    Returns
    -------
    gpd.GeoDataFrame
        GeoDataFrame containing the data from the file.

    Raises
    ------
    ValueError
        If the file type is not supported.

    """
    file = Path(file)
    if file.suffix in {".shp", ".gpkg"}:
        return gpd.read_file(file, **kwargs)
    elif file.suffix in {".parquet", ".geoparquet"}:
        return gpd.read_parquet(file, **kwargs)
    else:
        raise ValueError(f"File type {file.suffix} is not supported by geopandas.")


def csv_to_parquet(
    file: Union[str, Path], out_file: Union[str, Path] = None, **kwargs
) -> None:
    """
    Convert csv table to parquet.

    Parameters
    ----------
    file : Union[str, Path]
        Path to csv file to convert.
    out_file : Union[str, Path], optional
        Path to parquet file to be written. If not provided it will use the path of
        'file'.
    **kwargs
        pandas.read_csv kwargs. See pandas.read_csv documentation.

    Raises
    ------
    TypeError
        If 'file' is not a csv file
    """
    file = Path(file)
    if file.suffix != ".csv":
        raise TypeError(f"File must be a csv file, but a {file.suffix}-file was given")

    df = pd.read_csv(file, **kwargs)
    if out_file is None:
        df.to_parquet(file.parent / (file.stem + ".parquet"))
    else:
        df.to_parquet(out_file)


def excel_to_parquet(
    file: Union[str, Path], out_file: Union[str, Path] = None, **kwargs
) -> None:
    """
    Convert excel table to parquet.

    Parameters
    ----------
    file : Union[str, Path]
        Path to excel file to convert.
    out_file : Union[str, Path], optional
        Path to parquet file to be written. If not provided it will use the path of
        'file'.
    **kwargs
        pandas.read_excel kwargs. See pandas.read_excel documentation.

    Raises
    ------
    TypeError
        If 'file' is not an xlsx or xls file.
    """
    file = Path(file)
    if file.suffix not in [".xlsx", ".xls"]:
        raise TypeError(
            f"File must be an excel file, but a {file.suffix}-file was given"
        )

    df = pd.read_excel(file, **kwargs)
    if out_file is None:
        df.to_parquet(file.parent / (file.stem + ".parquet"))
    else:
        df.to_parquet(out_file)


def get_path_iterable(path: Path, wildcard: str = "*"):
    if path.is_file():
        return [path]
    elif path.is_dir():
        return path.glob(wildcard)
    else:
        raise TypeError("Given path is not a file or a folder")


def safe_float(number):
    try:
        return float(number)
    except ValueError:
        return None


def dataframe_to_geodataframe(
    df: pd.DataFrame, x_col_label: str = "x", y_col_label: str = "y", crs: int = None
) -> gpd.GeoDataFrame:
    """
    Take a dataframe with columns that indicate x and y coordinates and use these to
    turn the dataframe into a geopandas GeoDataFrame with a geometry column that
    contains shapely Point geometries.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe with columns for x and y coordinates.
    x_col_label : str
        Label of the x-coordinate column, default x-coordinate column label is 'x'.
    y_col_label : str
        Label of the y-coordinate column, default y-coordinate column label is 'y'.
    crs : int
        EPSG number as integer.

    Returns
    -------
    gpd.GeoDataFrame
        GeoDataFrame with point geometries in addition to input dataframe data.

    Raises
    ------
    IndexError
        If input dataframe does not have a valid column for 'x' or 'y'.
    """
    points = [gmt.Point([x, y]) for x, y in zip(df[x_col_label], df[y_col_label])]
    gdf = gpd.GeoDataFrame(df, geometry=points, crs=crs)
    return gdf


def save_pickle(data: Any, path: str | Path, **kwargs) -> None:
    """
    Save object as a pickle file (.pkl, .pickle) using Pandas to_pickle.

    Parameters
    ----------
    data : Any
        Object to save.
    path : str | Path
        Path to pickle file.
    **kwargs
            pd.to_pickle kwargs. See relevant Pandas documentation.

    """
    pd.to_pickle(data, path, **kwargs)


def _to_geopackage(
    data: gpd.GeoDataFrame, outfile: str | Path, error_note: str, **kwargs
):
    """
    Helper to add GeoST specific information if a Pyogrio error is raised when data is
    exported to Geopackage but contians invalid column names.

    Raises
    ------
    e
        Pyogrio error with added GeoST information when data has invalid column names.
    """
    try:
        data.to_file(outfile, **kwargs)
    except FieldError as e:
        e.add_note(f"Invalid column name in {error_note}, cannot write GPKG.")
        raise e


def create_connection(database: str | Path):
    """
    Create a database connection to a SQLite database.

    Parameters
    ----------
    database: string
        Path/url/etc. to the database to create the connection to.

    Returns
    -------
    conn : sqlite3.Connection
        Connection object or None.

    """
    conn = None
    try:
        conn = sqlite3.connect(database)
    except sqlite3.Error as e:
        print(e)

    return conn


def check_geometry_instance(
    geometry: str | Path | gpd.GeoDataFrame | Geometry | list[Geometry],
) -> gpd.GeoDataFrame:
    """
    Check if the input geometry is a valid type and convert it to a GeoDataFrame if
    necessary.

    Parameters
    ----------
    geometry : str | Path | gpd.GeoDataFrame | Geometry | list[Geometry]
        The geometry to check.

    Returns
    -------
    gpd.GeoDataFrame
        An instance of a geopandas geodataframe
    """
    if isinstance(geometry, str | Path):
        gdf = _geopandas_read(Path(geometry))
    elif isinstance(geometry, gpd.GeoDataFrame):
        gdf = geometry
    elif isinstance(geometry, gmt.base.BaseGeometry):
        gdf = gpd.GeoDataFrame(geometry=[geometry])
    elif isinstance(geometry, list):
        gdf = gpd.GeoDataFrame(geometry=geometry)

    # Make sure you don't get a view of gdf returned
    gdf = gdf.copy()

    return gdf



import re

import numpy as np
import pytest
import pyvista as pv
import xarray as xr
from numpy.testing import assert_array_equal

from geost.models.basemodels import VoxelModel


@pytest.fixture
def point_shapefile(point_header_gdf, tmp_path):
    shapefile = tmp_path / "point_shapefile.shp"
    point_header_gdf.to_file(shapefile)
    return shapefile


@pytest.fixture
def point_parquet(point_header_gdf, tmp_path):
    parquet = tmp_path / "point_shapefile.geoparquet"
    point_header_gdf.to_parquet(parquet)
    return parquet


@pytest.fixture
def voxelmodel_netcdf(xarray_dataset, tmp_path):
    outfile = tmp_path / "voxelmodel.nc"
    xarray_dataset.to_netcdf(outfile)
    return outfile


class TestVoxelModel:
    @pytest.mark.unittest
    def test_from_netcdf(self, voxelmodel_netcdf):
        model = VoxelModel.from_netcdf(voxelmodel_netcdf)
        assert isinstance(model, VoxelModel)

        model = VoxelModel.from_netcdf(
            voxelmodel_netcdf, data_vars=["strat"], bbox=(1, 1, 3, 3), lazy=False
        )
        assert isinstance(model, VoxelModel)
        assert model.horizontal_bounds == (1, 1, 3, 3)
        assert list(model.variables) == ["strat"]

    @pytest.mark.unittest
    def test_initialize(self, xarray_dataset):
        model = VoxelModel(xarray_dataset)
        assert isinstance(model, VoxelModel)
        # TODO: Make sure an input DataArray becomes a Dataset which is needed by all methods.

    @pytest.mark.unittest
    def test_attributes(self, voxelmodel):
        assert voxelmodel.sizes == {"y": 4, "x": 4, "z": 4}
        assert voxelmodel.shape == (4, 4, 4)
        assert voxelmodel.resolution == (1, 1, 0.5)
        assert voxelmodel.horizontal_bounds == (0, 0, 4, 4)
        assert voxelmodel.vertical_bounds == (-2, 0)
        assert voxelmodel.crs == 28992
        assert_array_equal(voxelmodel.variables, ["strat", "lith"])
        assert voxelmodel.xmin == 0
        assert voxelmodel.ymin == 0
        assert voxelmodel.xmax == 4
        assert voxelmodel.ymax == 4
        assert voxelmodel.zmin == -2
        assert voxelmodel.zmax == 0

    @pytest.mark.unittest
    def test_sel(self, voxelmodel):
        ## Select exact coordinates
        selected = voxelmodel.sel(x=[1.5, 2.5])
        assert isinstance(selected, VoxelModel)
        assert selected.shape == (4, 2, 4)

        ## Other selections
        selected = voxelmodel.sel(x=[1.7, 2.3], method="nearest")
        assert selected.shape == (4, 2, 4)
        assert_array_equal(selected["x"], [1.5, 2.5])

        selected = voxelmodel.sel(x=slice(0.1, 2.5))
        assert selected.shape == (4, 3, 4)
        assert_array_equal(selected["x"], [0.5, 1.5, 2.5])

    @pytest.mark.unittest
    def test_isel(self, voxelmodel):
        selected = voxelmodel.isel(x=[0, 2])
        assert isinstance(selected, VoxelModel)
        assert selected.shape == (4, 2, 4)
        assert_array_equal(selected["x"], [0.5, 2.5])

        selected = voxelmodel.isel(x=slice(0, 2))
        assert selected.shape == (4, 2, 4)
        assert_array_equal(selected["x"], [0.5, 1.5])

    @pytest.mark.unittest
    def test_select_with_points(self, voxelmodel, borehole_collection):
        points = borehole_collection.header.gdf
        select = voxelmodel.select_with_points(points)
        assert isinstance(select, xr.Dataset)
        assert select.sizes == {"idx": 4, "z": 4}
        assert_array_equal(select["idx"], [0, 1, 2, 4])
        assert_array_equal(select.data_vars, ["strat", "lith"])

    @pytest.mark.unittest
    @pytest.mark.parametrize("file", ["point_shapefile", "point_parquet"])
    def test_select_with_points_from_file(self, voxelmodel, file, request):
        file = request.getfixturevalue(file)
        select = voxelmodel.select_with_points(file)
        assert isinstance(select, xr.Dataset)

    @pytest.mark.unittest
    def test_thickness_map_single_condition(self, voxelmodel):
        thickness_lith1 = voxelmodel.get_thickness(voxelmodel["lith"] == 1)
        thickness_lith2 = voxelmodel.get_thickness(voxelmodel["lith"] == 2)
        thickness_lith3 = voxelmodel.get_thickness(voxelmodel["lith"] == 3)
        assert isinstance(thickness_lith1, xr.DataArray)
        assert isinstance(thickness_lith2, xr.DataArray)
        assert isinstance(thickness_lith3, xr.DataArray)
        assert_array_equal(
            thickness_lith1,
            np.array(
                [
                    [0.5, 1.0, 1.5, 1.0],
                    [1.0, 1.0, 1.5, 0.5],
                    [0.5, 1.0, 1.0, 0.5],
                    [0.5, 1.5, 0.5, 0.5],
                ]
            ),
        )
        assert_array_equal(
            thickness_lith2,
            np.array(
                [
                    [1.0, 0.5, 0.5, 0.5],
                    [1.0, 1.0, 0.5, 1.0],
                    [1.0, 0.5, 1.0, 1.5],
                    [1.5, 0.5, 1.0, 1.0],
                ]
            ),
        )
        assert_array_equal(
            thickness_lith3,
            np.array(
                [
                    [0.5, 0.5, 0.0, 0.5],
                    [0.0, 0.0, 0.0, 0.5],
                    [0.5, 0.5, 0.0, 0.0],
                    [0.0, 0.0, 0.5, 0.5],
                ]
            ),
        )

    @pytest.mark.unittest
    def test_thickness_map_extra_conditions(self, voxelmodel):
        thickness_lith2_strat2 = voxelmodel.get_thickness(
            ((voxelmodel["lith"] == 2) & (voxelmodel["strat"] == 2))
        )
        thickness_lith2_zrange = voxelmodel.get_thickness(
            voxelmodel["lith"] == 2, depth_range=(-1, 0)
        )
        assert isinstance(thickness_lith2_strat2, xr.DataArray)
        assert isinstance(thickness_lith2_zrange, xr.DataArray)
        assert_array_equal(
            thickness_lith2_strat2,
            np.array(
                [
                    [1.0, 0.5, 0.5, 0.5],
                    [1.0, 1.0, 0.5, 1.0],
                    [1.0, 0.5, 1.0, 1.5],
                    [1.5, 0.5, 1.0, 1.0],
                ]
            ),
        )
        assert_array_equal(
            thickness_lith2_zrange,
            np.array(
                [
                    [0.5, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.5],
                    [0.5, 0.0, 0.0, 0.5],
                    [0.5, 0.0, 0.0, 0.5],
                ]
            ),
        )

    @pytest.mark.unittest
    def test_to_pyvista_structured(self, voxelmodel):
        vms_single_var = voxelmodel.to_pyvista_grid(data_vars=["strat"])
        vms_multi_var = voxelmodel.to_pyvista_grid()
        assert isinstance(vms_single_var, pv.ImageData)
        assert isinstance(vms_multi_var, pv.ImageData)
        assert vms_single_var.n_points == 125
        assert vms_single_var.n_cells == 64
        assert vms_single_var.n_arrays == 1
        assert vms_multi_var.n_points == 125
        assert vms_multi_var.n_cells == 64
        assert vms_multi_var.n_arrays == 2

    @pytest.mark.unittest
    def test_to_pyvista_unstructured(self, voxelmodel):
        vmu_single_var = voxelmodel.to_pyvista_grid(
            data_vars=["strat"], structured=False
        )
        vmu_multi_var = voxelmodel.to_pyvista_grid(structured=False)
        assert isinstance(vmu_single_var, pv.UnstructuredGrid)
        assert isinstance(vmu_multi_var, pv.UnstructuredGrid)
        assert vmu_single_var.n_points == 512
        assert vmu_single_var.n_cells == 64
        assert vmu_single_var.n_arrays == 1
        assert vmu_multi_var.n_points == 512
        assert vmu_multi_var.n_cells == 64
        assert vmu_multi_var.n_arrays == 2

    @pytest.mark.unittest
    def test_to_pyvista_unstructured_problematic_dims(self, voxelmodel):
        # Wrong order of dimensions leads to automatic transposing, not an error!
        vmu_wrong_order = voxelmodel.to_pyvista_grid(structured=False)
        assert isinstance(vmu_wrong_order, pv.UnstructuredGrid)
        assert vmu_wrong_order.n_points == 512
        assert vmu_wrong_order.n_cells == 64
        assert vmu_wrong_order.n_arrays == 2

        # Missing z-dimension leads to an error and no file is created.
        voxelmodel.ds = voxelmodel.ds.drop_vars("z")
        with pytest.raises(Exception) as error_info:
            voxelmodel.to_pyvista_grid()
        assert error_info.errisinstance(ValueError)
        assert error_info.match(
            re.escape(
                "Dataset must contain 'z' dimension. Make sure that this "
                "spatial dimension exists in the dataset or if it has a different "
                "name use xarray.Dataset.rename() to rename the corresponding "
                "dimension to 'z'."
            )
        )



from typing import NamedTuple, Union


class RDCoord(NamedTuple):
    x: Union[int, float]
    y: Union[int, float]
    epsg: int


class DDCoord(NamedTuple):
    lat: Union[int, float]
    lon: Union[int, float]
    epsg: int


def safe_coerce(variable, dtype):
    try:
        variable = dtype(variable)
    except ValueError:
        Warning(f"Failed to coerce {variable} to {dtype}. Please check parsed data.")

    return variable



from dataclasses import dataclass


@dataclass
class Cpt:
    nr: str = None
    x: int | float = None
    y: int | float = None
    z: int | float = None
    enddepth: int | float = None
    gefid = None
    ncolumns = None
    columninfo: dict = None
    companyid = None
    filedate = None
    fileowner = None
    lastscan = None
    procedurecode = None  # mandatory if gefid is 1, 0, 0
    reportcode = None  # this or procedurecode is mandatory if gefid > 1, 1, 0
    projectid = None
    measurementtext: dict = None
    columnvoid: dict = None
    columnminmax = None
    dataformat = None
    measurementvars: dict = None
    recordseparator = None
    reportdataformat = None
    specimenvars = None
    startdate = None
    starttime = None
    coord_system = None
    reference_system = None

    def __post_init__(self):
        pass



from pathlib import Path

import geopandas as gpd
import pandas as pd
import pytest
import xarray as xr

import geost
from geost.base import BoreholeCollection, CptCollection
from geost.bro import GeoTop


@pytest.mark.unittest
def test_boreholes_usp():
    boreholes = geost.data.boreholes_usp()
    assert isinstance(boreholes, BoreholeCollection)
    boreholes = geost.data.boreholes_usp(pandas=True)
    assert isinstance(boreholes, pd.DataFrame)


@pytest.mark.unittest
def test_cpts_usp():
    cpts = geost.data.cpts_usp()
    assert isinstance(cpts, CptCollection)
    cpts = geost.data.cpts_usp(pandas=True)
    assert isinstance(cpts, pd.DataFrame)


@pytest.mark.unittest
def test_geotop_usp():
    gtp = geost.data.geotop_usp()
    assert isinstance(gtp, GeoTop)
    gtp = geost.data.geotop_usp(xarray=True)
    assert isinstance(gtp, xr.Dataset)


@pytest.mark.xfail(reason="Will work if branch 'feature/xml` is merged into main.")
@pytest.mark.unittest
def test_bhrg_bro():
    bhrg = geost.data.bhrg_bro()
    assert isinstance(bhrg, Path)
    assert bhrg.isfile()


@pytest.mark.xfail(reason="Will work if branch 'feature/xml` is merged into main.")
@pytest.mark.unittest
def test_dike_section():
    dike = geost.data.dike_section()
    assert isinstance(dike, gpd.GeoDataFrame)



import importlib

import geopandas as gpd
import pandas as pd
import pooch
import xarray as xr

from geost.bro import GeoTop
from geost.io.read import read_borehole_table, read_cpt_table

REGISTRY = pooch.create(
    path=pooch.os_cache("geost"),
    base_url="https://github.com/Deltares-research/geost/raw/main/data/",
    version=None,
    env="GEOST_DATA_DIR",
)
REGISTRY.load_registry(importlib.resources.files("geost.data") / "registry.txt")


def boreholes_usp(pandas=False):
    """
    Return a :class:`~geost.base.BoreholeCollection` with a selection of DINOloket boreholes
    in the area of the Utrecht Science Park to use in GeoST tutorials.

    Parameters
    ----------
    pandas : bool, optional
        If True, read the boreholes into a `Pandas DataFrame`. The default is False, this
        returns a :class:`~geost.base.BoreholeCollection`.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection`

    """
    filename = REGISTRY.fetch("boreholes_usp.parquet")
    if pandas:
        return pd.read_parquet(filename)
    else:
        return read_borehole_table(filename)


def cpts_usp(pandas=False):
    """
    Return a :class:`~geost.base.CptCollection` with a selection of BROloket CPTs in the
    area of the Utrecht Science Park to use in GeoST tutorials.

    Parameters
    ----------
    pandas : bool, optional
        If True, read the boreholes into a `Pandas DataFrame`. The default is False, this
        returns a :class:`~geost.base.CptCollection`.

    Returns
    -------
    :class:`~geost.base.CptCollection`

    """
    filename = REGISTRY.fetch("cpts_usp.parquet")
    if pandas:
        return pd.read_parquet(filename)
    else:
        return read_cpt_table(filename)


def geotop_usp(xarray=False):
    """
    Return a :class:`~geost.bro.GeoTop` instance in the area of the Utrecht Science Park
    to use in GeoST tutorials.

    Parameters
    ----------
    xarray : bool, optional
        If True, read the GeoTOP data as an `xarray.Dataset`. The default is False, this
        returns a :class:`~geost.bro.GeoTop`.

    Returns
    -------
    :class:`~geost.bro.GeoTop`

    """
    filename = REGISTRY.fetch("geotop_usp.nc")
    if xarray:
        return xr.open_dataset(filename)
    else:
        return GeoTop.from_netcdf(filename)


def bhrg_bro():
    """
    Download a BRO BHR-G XML file from the GeoST data registry and return a `Pathlib.Path`
    object pointing to the file for tutorials explaining GeoST's XML parsing.

    Returns
    -------
    Pathlib.Path
        A Pathlib.Path object pointing to the downloaded XML file.

    """
    return REGISTRY.fetch("bhrg_bro.xml")


def dike_section():
    """
    Download a dike section geoparquet file from the GeoST data registry and return a
    Geopandas `GeoDataFrame` of the dike section for tutorials.

    Returns
    -------
    GeoDataFrame
        A Geopandas GeoDataFrame object containing the dike section data.

    """
    filename = REGISTRY.fetch("dike_section.geoparquet")
    return gpd.read_parquet(filename)



import numpy as np
import pytest
import xarray as xr
from numpy.testing import assert_array_equal

from geost.bro.bro_utils import coordinates_to_cellcenters, flip_ycoordinates


@pytest.fixture
def dataset():
    coords = np.arange(3)
    return xr.DataArray(
        np.ones((3, 3, 3)),
        coords={"y": coords, "x": coords, "z": [0.5, 1.5, 2.5]},
        dims=("y", "x", "z"),
    )


@pytest.mark.unittest
def test_coordinates_to_cellcenters(dataset: xr.DataArray):
    cellsize, dz = 1, 1
    dataset = coordinates_to_cellcenters(dataset, cellsize, dz)
    assert_array_equal(dataset["x"], [0.5, 1.5, 2.5])
    assert_array_equal(dataset["y"], [0.5, 1.5, 2.5])
    assert_array_equal(dataset["z"], [1, 2, 3])


@pytest.mark.unittest
def test_flip_ycoordinates(dataset: xr.DataArray):
    dataset = flip_ycoordinates(dataset)
    assert_array_equal(dataset["y"], [2, 1, 0])
    assert dataset.dims == ("y", "x", "z")



"""
This code is part of the core functionality of Deltares DataFusionTools. The 'Data'
class is used to represent all spatial data, including boreholes and CPT's. In order to
avoid a dependency on DataFusionTools (and its many dependencies that geost doesn't
need), the code was copied here to allow for the export of geost objects to
DataFusionTools Data objects.

source:

https://bitbucket.org/DeltaresGEO/datafusiontools/src/master/

DataFusionTools is released under GNU GPL v3
"""

from dataclasses import dataclass
from typing import List, Optional, Union

import numpy as np
import pandas as pd


@dataclass
class Geometry:
    x: float
    y: float
    z: Optional[float] = None
    label: Optional[float] = None


@dataclass
class Variable:
    label: str
    value: np.array


@dataclass
class Data:
    """
    Class that contains values of one location

    :param location: The location the data
    :param variables: Variables that are contain in the data
    :param independent_variable: Independent variable that is connected to all the data
    """

    location: Geometry
    independent_variable: Union[Variable, None] = None
    variables: List[Variable] = []

    def get_variable(self, name: str):
        """
        Function that returns variable based on its name
        """
        for variable in self.variables:
            if variable.label == name:
                return variable
        return None

    def update_variable(self, name: str, value: np.ndarray):
        """
        Function that updates variable
        """
        index = None
        for counter, variable in enumerate(self.variables):
            if name == variable.label:
                index = counter
                continue
        if index is None:
            raise ValueError(f"Name provided {name} is defined in this class instance.")
        self.variables[index].value = value

    @property
    def variables(self):
        return self._variables

    @variables.setter
    def variables(self, values: List[Variable]):
        if self.independent_variable is not None:
            for variable in values:
                if len(variable.value) != len(self.independent_variable.value):
                    raise ValueError(
                        f"Length of variable {variable.label} is not the same as that ",
                        "of the independent variable.",
                    )
        self._variables = values


########################################################################################
################################# geost ADDITIONS ######################################
########################################################################################


def export_to_dftgeodata(data, columns, encode=True):
    if encode:
        data_encoded = pd.get_dummies(data[columns])
        data.drop(columns=columns, inplace=True)
        columns = data_encoded.columns
        data = pd.concat([data, data_encoded], axis=1)

    objects = data.groupby("nr")
    geodataclasses = []

    for _, obj in objects:
        variables = []

        location = Geometry(
            obj["x"].iloc[0], obj["y"].iloc[0], z=obj["surface"].iloc[0]
        )

        independent_var_depth = Variable(
            value=(obj["top"] - (obj["top"] - obj["bottom"]) / 2).values, label="height"
        )

        for col in columns:
            variables.append(Variable(value=obj[col].values, label=col))

        geodataclasses.append(Data(location, independent_var_depth, variables))

    return geodataclasses



import numpy as np
import pandas as pd
import xarray as xr

from geost.base import Collection, DiscreteData, LayeredData
from geost.models import VoxelModel
from geost.models.model_utils import label_consecutive_2d


def add_voxelmodel_variable(
    collection: Collection, model: VoxelModel, variable: str
) -> Collection:
    """
    Add a information from a variable of a `VoxelModel` instance as a column to the data
    of a `BoreholeCollection` or `CptCollection` instance. This checks for each data object
    in the Collection instance in which voxel stack the object is located and adds the
    relevant layer boundaries of the variable to the data object of the Collection based
    on depth.

    Note
    ----
    If the variable name is already present in the columns of the data attribute of the
    collection, the present column is overwritten. To avoid this, rename the variable
    before either in collection.data or in the voxelmodel.

    Parameters
    ----------
    collection : Collection
        Any GeoST `Collection` object such as a :class:`~geost.base.BoreholeCollection`
        or :class:`~geost.base.CptCollection` to add the `VoxelModel` variable to.
    model : :class:`~geost.modelbase.VoxelModel`
        `VoxelModel` instance to add the information from to the `Collection` instance.
    variable : str
        Name of the variable in the `VoxelModel` to add.

    Returns
    -------
    Collection
        New `Collection` (same type as the input) object with the variable added to the
        data object of the `Collection`.

    Raises
    ------
    NotImplementedError
        Raises error for geost data object types that are not yet implemented such as
        :class:`~geost.base.LineData` (future developments).

    """
    if variable in collection.data.df.columns:
        collection.data.df.drop(columns=[variable], inplace=True)

    var_select = model.select_with_points(collection.header.gdf)[variable]
    nrs = collection.header["nr"].loc[var_select["idx"]]

    _, _, dz = model.resolution

    var_df = _reduce_to_top_bottom(var_select, dz)
    var_df.rename(columns={"values": variable}, inplace=True)
    var_df["nr"] = nrs.loc[var_df["nr"]].values

    if isinstance(collection.data, LayeredData):
        result = _add_to_layered(collection.data, var_df)
    elif isinstance(collection.data, DiscreteData):
        result = _add_to_discrete(collection.data, var_df)
    else:
        raise NotImplementedError(
            "Other datatypes than LayeredData or DiscreteData not implemented yet."
        )

    return result.to_collection()


def _reduce_to_top_bottom(da: xr.DataArray, dz: int | float) -> pd.DataFrame:
    """
    Helper to reduce the selection DataArray from `VoxelModel.select_with_points` to a
    DataFrame containing "idx", "top" and "bottoms" of relevant layer boundaries.

    Parameters
    ----------
    da : xr.DataArray
        Selection result of `VoxelModel.select_with_points`.
    dz : int | float
        Vertical size of voxels in the VoxelModel.

    Returns
    -------
    pd.DataFrame

    """
    layer_ids = label_consecutive_2d(da.values, axis=1)
    df = pd.DataFrame(
        {
            "nr": np.repeat(da["idx"], da.sizes["z"]),
            "bottom": np.tile(da["z"] - (0.5 * dz), da.sizes["idx"]),
            "layer": layer_ids.ravel(),
            "values": da.values.ravel(),
        }
    )
    reduced = df.pivot_table(
        index=["nr", "layer", "values"], values="bottom", aggfunc="min"
    ).reset_index()
    return reduced.drop(columns=["layer"])


def _add_to_layered(data: LayeredData, variable: pd.DataFrame) -> LayeredData:
    """
    Helper function for `add_voxelmodel_variable` to combine a voxelmodel variable in
    layered data form. This joins the DataFrames of the LayeredData instance and the
    variable from the voxelmodel to add and updates the "top" and "bottom" columns in the
    LayeredData.

    Parameters
    ----------
    data : LayeredData
        LayeredData instance to add the voxelmodel variable to.
    variable : pd.DataFrame
        DataFrame with the voxelmodel variable to add. Contains columns ["nr", "bottom",
        variable] where variable is for example "strat".

    Returns
    -------
    LayeredData
        New instance of `LayeredData` with the added variable and corrected top and bottom
        depths.

    """
    variable = variable.merge(
        data[["nr", "surface"]].drop_duplicates(), on="nr", how="left"
    )
    variable["bottom"] = variable["surface"] - variable["bottom"]

    result = (
        pd.concat([data.df, variable])
        .sort_values(by=["nr", "bottom", "top"])
        .reset_index(drop=True)
    )
    nr = result["nr"]
    result = pd.concat([nr, result.groupby("nr").bfill()], axis=1)
    result.drop_duplicates(subset=["nr", "bottom"], inplace=True)
    result = _reset_tops(result)
    result.dropna(subset=["top"], inplace=True)
    return LayeredData(result)


def _reset_tops(layered_df: pd.DataFrame) -> pd.DataFrame:
    """
    Helper function for `_add_to_layered` to reset the top depths after stratigraphic
    layer boundaries have been added and backfilling has been done. This changes tops into
    bottoms of the row above in each borehole if tops are lower than bottoms. It also checks
    if bottoms are less than 0 and resets these layers to 0. Then it removes layers with
    a thickness of 0.

    Parameters
    ----------
    layered_df : pd.DataFrame
        Pandas DataFrame with layered data containing tops and bottoms.

    Returns
    -------
    pd.DataFrame
        DataFrame with the tops reset.
    """
    bottom_shift_down = layered_df["bottom"].shift()
    nr_shift_down = layered_df["nr"].shift()

    layered_df.loc[
        (layered_df["top"] < bottom_shift_down) & (layered_df["nr"] == nr_shift_down),
        "top",
    ] = bottom_shift_down

    layered_df.loc[layered_df["bottom"] < 0, "bottom"] = 0

    return layered_df[layered_df["top"] - layered_df["bottom"] != 0].reset_index(
        drop=True
    )


def _add_to_discrete(data: DiscreteData, variable: pd.DataFrame) -> DiscreteData:
    """
    Helper function for `add_voxelmodel_variable` to combine a voxelmodel variable in
    discrete data form. This joins the DataFrames of the DiscreteData instance and the
    variable from the voxelmodel to return a new instance of DiscreteData.

    Parameters
    ----------
    data : DiscreteData
        DiscreteData instance to add the voxelmodel variable to.
    variable : pd.DataFrame
        DataFrame with the voxelmodel variable to add. Contains columns ["nr", "bottom",
        variable] where variable is for example "strat".

    Returns
    -------
    DiscreteData
        New instance of `DiscreteData` with the added variable.

    """
    variable.rename(columns={"bottom": "depth"}, inplace=True)
    variable = variable.merge(
        data[["nr", "surface"]].drop_duplicates(), on="nr", how="left"
    )
    variable["depth"] = variable["surface"] - variable["depth"]

    result = (
        pd.concat([data.df, variable])
        .sort_values(by=["nr", "depth"])
        .reset_index(drop=True)
    )
    nr = result["nr"]
    result = pd.concat([nr, result.groupby("nr").bfill()], axis=1)
    result.drop_duplicates(subset=["nr", "depth"], inplace=True)
    result.dropna(subset=["end"], inplace=True)
    return DiscreteData(result)



from geost.io.xml import resolvers

BRO = {
    "payload_root": "dispatchDocument",
    "nr": {"xpath": "brocom:broId"},
    "location": {
        "xpath": "deliveredLocation/cptcommon:location/gml:pos",
        "resolver": resolvers.parse_coordinates,
        "el-attr": "text",
    },
    "crs": {
        "xpath": "deliveredLocation/cptcommon:location",
        "resolver": resolvers.parse_crs,
    },
    "surface": {
        "xpath": "deliveredVerticalPosition/cptcommon:offset",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "vertical_datum": {
        "xpath": "deliveredVerticalPosition/cptcommon:verticalDatum",
        "el-attr": "text",
    },
    "predrilled_depth": {
        "xpath": "conePenetrometerSurvey/cptcommon:trajectory/cptcommon:predrilledDepth",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "end": {
        "xpath": "conePenetrometerSurvey/cptcommon:trajectory/cptcommon:finalDepth",
        "resolver": resolvers.safe_float,
        "el-attr": "text",
    },
    "data": {
        "xpath": "conePenetrometerSurvey",
        "resolver": resolvers.process_cpt_data,
    },
}

schemas = {"BRO": BRO}



import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal
from shapely.geometry import LineString

from geost.models import model_utils as mut


@pytest.fixture
def line():
    return LineString([[1, 1], [3, 3]])


@pytest.fixture
def array_2d():
    return np.array(
        [
            [10, 10, 10, 10, 10],
            [10, 20, 10, 20, 10],
            [20, 10, 20, 20, 10],
            [20, 20, 20, 10, 10],
        ]
    )


@pytest.mark.unittest
def test_interpolate_point(line):
    loc, x, y = mut._interpolate_point(line, 1)
    assert loc == 1
    assert np.isclose(x, 1.7071)
    assert np.isclose(y, 1.7071)


@pytest.mark.unittest
def test_sample_along_line(xarray_dataset, line):
    sampled = mut.sample_along_line(xarray_dataset, line, nsamples=3)
    assert sampled.sizes == {"dist": 3, "z": 4}
    assert_array_almost_equal(sampled["dist"], [0.0, 1.414, 2.828], decimal=3)

    sampled = mut.sample_along_line(xarray_dataset, line, dist=0.5)
    assert sampled.sizes == {"dist": 6, "z": 4}
    assert_array_almost_equal(sampled["dist"], [0, 0.5, 1, 1.5, 2, 2.5])


@pytest.mark.unittest
def test_sample_with_coords(xarray_dataset, borehole_collection):
    coords = borehole_collection.header[["x", "y"]].values
    sampled = mut.sample_with_coords(xarray_dataset, coords)
    assert sampled.sizes == {"idx": 4, "z": 4}
    assert_array_equal(sampled["idx"], [0, 1, 2, 4])


@pytest.mark.unittest
def test_label_consecutive_2d(array_2d):
    result = mut.label_consecutive_2d(array_2d, axis=0)
    assert_array_equal(
        result, [[0, 0, 0, 0, 0], [0, 1, 0, 1, 0], [1, 2, 1, 1, 0], [1, 3, 1, 2, 0]]
    )
    assert result.dtype == array_2d.dtype

    result = mut.label_consecutive_2d(array_2d, axis=1)
    assert_array_equal(
        result, [[0, 0, 0, 0, 0], [0, 1, 2, 3, 4], [0, 1, 2, 2, 3], [0, 0, 0, 1, 1]]
    )
    assert result.dtype == array_2d.dtype



class ValidationWarning(UserWarning):
    """Warning for when validation fails."""

    pass


class AlignmentWarning(UserWarning):
    """Warning for when header and data are not aligned."""

    pass



from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import xarray as xr
from numpy.testing import assert_allclose

from geost import spatial
from geost.utils import dataframe_to_geodataframe


class TestSpatialUtils:
    @pytest.fixture
    def raster(self):
        x_coors = np.arange(1, 4)
        y_coors = np.arange(4, 1, -1)
        data = np.ones((3, 3))
        array = xr.DataArray(data, {"x": x_coors, "y": y_coors})
        return array

    @pytest.fixture
    def invalid_raster(self):
        x_coors = np.arange(1, 4)
        y_coors = np.arange(4, 1, -1)
        data = np.ones((3, 3))
        array = xr.DataArray(data, {"invalid_x": x_coors, "invalid_y": y_coors})
        return array

    @pytest.fixture
    def dataframe_with_coordinates(self):
        dataframe = pd.DataFrame({"x": [1, 2, 3, 4, 5], "y": [5, 4, 3, 2, 1]})
        return dataframe

    @pytest.mark.unittest
    def test_dataframe_to_geodataframe(self, dataframe_with_coordinates):
        gdf = dataframe_to_geodataframe(dataframe_with_coordinates, crs=28992)
        assert isinstance(gdf["geometry"].dtype, gpd.array.GeometryDtype)

    # @pytest.mark.unittest
    # def test_check_geometry_instance(self, point_header_gdf):
    #     point_header_gdf.to_parquet("temp_file.geoparquet")
    #     point_header_gdf.to_file("temp_file.gpkg")
    #     point_header_gdf_geoparquet = "temp_file.geoparquet"
    #     point_header_gdf_geopackage = "temp_file.gpkg"
    #     gdf_gdf = spatial.check_geometry_instance(point_header_gdf)
    #     gdf_geoparquet_gdf = spatial.check_geometry_instance(
    #         point_header_gdf_geoparquet
    #     )
    #     gdf_geopackage_gdf = spatial.check_geometry_instance(
    #         point_header_gdf_geopackage
    #     )
    #     assert isinstance(gdf_gdf, gpd.GeoDataFrame)
    #     assert isinstance(gdf_geoparquet_gdf, gpd.GeoDataFrame)
    #     assert isinstance(gdf_geopackage_gdf, gpd.GeoDataFrame)
    #     Path("temp_file.geoparquet").unlink()
    #     Path("temp_file.gpkg").unlink()

    @pytest.mark.unittest
    def test_check_and_coerce_crs(self, point_header_gdf):
        with pytest.warns(UserWarning):
            referenced_gdf = spatial.check_and_coerce_crs(point_header_gdf, 28992)
            converted_referenced_gdf = spatial.check_and_coerce_crs(
                referenced_gdf, 32631
            )
        assert referenced_gdf.crs == "epsg:28992"
        assert converted_referenced_gdf.crs == "epsg:32631"

    @pytest.mark.unittest
    def test_get_raster_values(self, raster, dataframe_with_coordinates):
        x = dataframe_with_coordinates["x"].values
        y = dataframe_with_coordinates["y"].values
        raster_values = spatial.get_raster_values(x, y, raster)
        assert_allclose(raster_values, np.array([np.nan, 1.0, 1.0, np.nan, np.nan]))

    @pytest.mark.unittest
    def test_get_invalid_raster_values(
        self, invalid_raster, dataframe_with_coordinates
    ):
        # In this case the used raster does not have the correct labels for x and y
        # coordinates, which should raise an error.
        x = dataframe_with_coordinates["x"].values
        y = dataframe_with_coordinates["y"].values
        with pytest.raises(Exception) as error_info:
            spatial.get_raster_values(x, y, invalid_raster)
        assert error_info.errisinstance(TypeError)
        assert error_info.match(
            "The xr.DataArray to sample from does not have the "
            + "required 'x' and 'y' dimensions"
        )



import geopandas as gpd
import pandas as pd
import pytest
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_array_equal,
    assert_equal,
)
from shapely.geometry import LineString, Point, Polygon

from geost.base import LineHeader, PointHeader


class TestHeaders:
    @pytest.mark.unittest
    def test_init_header(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        assert "PointHeader instance containing 25 objects" in point_header.__repr__()

    @pytest.mark.unittest
    def test_single_get(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        point_header_sel = point_header.get("nr10")
        assert point_header_sel["nr"].iloc[0] == "nr10"

    @pytest.mark.unittest
    def test_multi_get(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        query = ["nr1", "nr5", "nr10", "nr15", "nr20"]
        point_header_sel = point_header.get(query)
        assert all([nr in point_header_sel["nr"].values for nr in query])

    @pytest.mark.unittest
    def test_change_horizontal_reference(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        point_header.change_horizontal_reference(32631)
        assert point_header.horizontal_reference == 32631
        assert_almost_equal(point_header["x"][0], 523402.3476207458)
        assert_almost_equal(point_header["y"][0], 5313544.160440822)

    @pytest.mark.unittest
    def test_change_vertical_reference(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        current_NAP_levels = point_header["surface"].copy()  # noqa: N806
        current_NAP_ends = point_header["end"].copy()  # noqa: N806
        # The offset between NAP (epsg:5709) and TAW (epsg:5710) is 2.28234 m.
        expected_TAW_levels = current_NAP_levels + 2.28234  # noqa: N806
        expected_TAW_ends = current_NAP_ends + 2.28234  # noqa: N806
        point_header.change_vertical_reference(5710)
        assert point_header.vertical_reference == 5710
        assert_allclose(point_header["surface"], expected_TAW_levels, atol=10e-4)
        assert_allclose(point_header["end"], expected_TAW_ends, atol=10e-4)

    @pytest.mark.unittest
    def test_select_within_bbox(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        point_header_sel = point_header.select_within_bbox(1, 1, 3, 3)
        point_header_sel_inverted = point_header.select_within_bbox(
            1, 1, 3, 3, invert=True
        )
        assert len(point_header_sel) == 9
        assert len(point_header_sel_inverted) == 16

    @pytest.mark.unittest
    def test_select_with_points(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        selection_points = [Point(1, 1), Point(4, 4), Point(1, 4), Point(4, 1)]
        selection_gdf = gpd.GeoDataFrame(
            {"id": [0, 1, 2, 3]}, geometry=selection_points
        )
        selected = point_header.select_with_points(selection_gdf, 1.1)
        assert len(selected) == 16

        inverted_selection = point_header.select_with_points(
            selection_gdf, 1.1, invert=True
        )
        assert len(inverted_selection) == 9

        # Select with Shapely object
        selected = point_header.select_with_points(selection_points[0], 1.1)
        assert len(selected) == 3

        # Select with iterable
        selected = point_header.select_with_points(selection_points, 1.1)
        assert len(selected) == 16

    @pytest.mark.unittest
    def test_select_with_lines(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        selection_lines = [LineString([[1, 1], [5, 5]]), LineString([[1, 5], [5, 1]])]
        selection_gdf = gpd.GeoDataFrame({"id": [0, 1]}, geometry=selection_lines)

        selected = point_header.select_with_lines(selection_gdf, 1)
        assert len(selected) == 21

        inverted_selection = point_header.select_with_lines(
            selection_gdf, 1, invert=True
        )
        assert len(inverted_selection) == 4

        # Select with Shapely object
        selected = point_header.select_with_lines(selection_lines[0], 1)
        assert len(selected) == 13

        # Select with iterable
        selected = point_header.select_with_lines(selection_lines, 1)
        assert len(selected) == 21

    @pytest.mark.unittest
    def test_select_within_polygons(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        selection_polygon = Polygon(((2, 1), (5, 4), (4, 5), (1, 2)))
        selection_gdf = gpd.GeoDataFrame({"id": [0]}, geometry=[selection_polygon])

        # Geodataframe based selection
        selected = point_header.select_within_polygons(selection_gdf)
        assert len(selected) == 3

        selected_inverted = point_header.select_within_polygons(
            selection_gdf, invert=True
        )
        assert len(selected_inverted) == 22

        selected_buffered = point_header.select_within_polygons(
            selection_gdf, buffer=0.1
        )
        assert len(selected_buffered) == 11

        selected_inverted_buffered = point_header.select_within_polygons(
            selection_gdf, buffer=0.1, invert=True
        )
        assert len(selected_inverted_buffered) == 14

        # Shapely Polygon based selection
        selected = point_header.select_within_polygons(selection_polygon)
        assert len(selected) == 3
        assert_array_equal(selected["nr"], ["nr7", "nr13", "nr19"])

        # Selection with Iterable
        selected = point_header.select_within_polygons([selection_polygon])
        assert len(selected) == 3

    @pytest.mark.unittest
    def test_select_by_depth(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        assert len(point_header.select_by_depth(top_min=10).gdf) == 16
        assert len(point_header.select_by_depth(top_max=10).gdf) == 10
        assert len(point_header.select_by_depth(end_min=-10).gdf) == 10
        assert len(point_header.select_by_depth(end_max=-10).gdf) == 16
        assert len(point_header.select_by_depth(top_min=10, top_max=15).gdf) == 6
        assert (
            len(point_header.select_by_depth(top_min=10, top_max=15, end_max=-12).gdf)
            == 4
        )

    @pytest.mark.unittest
    def test_select_by_length(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        assert len(point_header.select_by_length(min_length=10)) == 21
        assert len(point_header.select_by_length(max_length=30)) == 15
        assert len(point_header.select_by_length(min_length=10, max_length=30)) == 11

    @pytest.mark.unittest
    def test_get_area_labels(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        label_polygon = [Polygon(((2, 1), (5, 4), (4, 5), (1, 2)))]
        label_gdf = gpd.GeoDataFrame({"id": [1]}, geometry=label_polygon, crs=28992)
        # Return variant
        output = point_header.get_area_labels(label_gdf, "id")
        assert_almost_equal(output["id"].sum(), 11)
        # In-place variant
        point_header.get_area_labels(label_gdf, "id", include_in_header=True)
        assert_almost_equal(point_header["id"].sum(), 11)

    @pytest.mark.unittest
    def test_get_area_labels_multiple(self, point_header_gdf):
        point_header = PointHeader(point_header_gdf, "NAP")
        label_polygon = [Polygon(((2, 1), (5, 4), (4, 5), (1, 2)))]
        label_gdf = gpd.GeoDataFrame(
            {"id": [1], "col2": ["string_data"]}, geometry=label_polygon, crs=28992
        )
        # Return variant
        output = point_header.get_area_labels(label_gdf, ["id", "col2"])
        assert_almost_equal(output["id"].sum(), 11)
        assert_equal(output["col2"].value_counts()["string_data"], 11)
        # In-place variant
        point_header.get_area_labels(label_gdf, ("id", "col2"), include_in_header=True)
        assert_almost_equal(point_header["id"].sum(), 11)
        assert_equal(point_header["col2"].value_counts()["string_data"], 11)

    @pytest.mark.unittest
    def test_export_to_csv_mixin(self, point_header_gdf, tmp_path):
        point_header = PointHeader(point_header_gdf, "NAP")
        temp_file = tmp_path / r"pointheader.csv"
        point_header.to_csv(temp_file)
        assert temp_file.is_file()

    @pytest.mark.unittest
    def test_export_to_parquet_mixin(self, point_header_gdf, tmp_path):
        point_header = PointHeader(point_header_gdf, "NAP")
        temp_file = tmp_path / r"pointheader.parquet"
        point_header.to_parquet(temp_file)
        assert temp_file.is_file()

    @pytest.mark.unittest
    def test_export_to_shapefile_mixin(self, point_header_gdf, tmp_path):
        point_header = PointHeader(point_header_gdf, "NAP")
        temp_file = tmp_path / r"pointheader.shp"
        point_header.to_shape(temp_file)
        assert temp_file.is_file()

    @pytest.mark.unittest
    def test_export_to_geopackage_mixin(self, point_header_gdf, tmp_path):
        point_header = PointHeader(point_header_gdf, "NAP")
        temp_file = tmp_path / r"pointheader.gpkg"
        point_header.to_geopackage(temp_file)
        assert temp_file.is_file()

    @pytest.mark.unittest
    def test_export_to_geoparquet_mixin(self, point_header_gdf, tmp_path):
        point_header = PointHeader(point_header_gdf, "NAP")
        temp_file = tmp_path / r"pointheader.geoparquet"
        point_header.to_geoparquet(temp_file)
        assert temp_file.is_file()



from enum import StrEnum
from pathlib import Path

import geopandas as gpd
import pandas as pd

from geost.io import Geopackage


class CptTables(StrEnum):  # pragma: no cover
    geotechnical_cpt_survey = "geotechnical_cpt_survey"
    delivered_location = "delivered_location"
    bro_location = "bro_location"
    bro_point = "bro_point"
    delivered_vertical_position = "delivered_vertical_position"
    cone_penetrometer_survey = "cone_penetrometer_survey"
    trajectory = "trajectory"
    processing = "processing"
    cone_penetrometer = "cone_penetrometer"
    zero_load_measurement = "zero_load_measurement"
    parameters = "parameters"
    cone_penetration_test = "cone_penetration_test"
    cone_penetration_test_result = "cone_penetration_test_result"
    dissipation_test = "dissipation_test"
    dissipation_test_result = "dissipation_test_result"
    additional_investigation = "additional_investigation"
    removed_layer = "removed_layer"
    registration_history = "registration_history"
    nga_properties = "nga_properties"


class BroCptGeopackage:
    """
    Class to handle the Bro CPT Geopackage file for data selections and facilitate making
    combinations between the different data tables in the BRO CPT geopackage.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        GeoDataFrame containing the spatial locations of the CPT data. The "find key" per
        "bro_id" to each related tables in the Geopackage as index (index name: "fid").
    db : :class:`~geost.io.Geopackage`
        Geost Geopackage instance to handle the database connections and queries.
    """

    def __init__(self, gdf: gpd.GeoDataFrame, db: Geopackage):
        self.gdf = gdf
        self.db = db

    @classmethod
    def from_geopackage(cls, file: str | Path, **gpd_kwargs):
        """
        Create an instance of `BroCptGeopackage` from a GeoPackage file.

        Parameters
        ----------
        file : str or Path
            The path to the GeoPackage file.
        **gpd_kwargs : dict
            Additional keyword arguments to pass to `gpd.read_file`. The 'fid_as_index'
            argument is set to True by default to retain the index fordatabase selections.
            The 'layer' argument should not be passed as it is set internally and this will
            raise a ValueError.

        Returns
        -------
        `BroCptGeopackage`

        Raises
        ------
        ValueError
            If the 'layer' argument is passed in `gpd_kwargs`. This argument is set internally
            to set the `gdf` attribute and cannot be overwritten.

        """
        if "fid_as_index" not in gpd_kwargs:  # Needs to retain index for db selections
            gpd_kwargs["fid_as_index"] = True

        if "layer" in gpd_kwargs:
            raise ValueError("Layer cannot be passed as a Geopandas keyword argument.")

        gdf = gpd.read_file(file, layer=CptTables.geotechnical_cpt_survey, **gpd_kwargs)
        db = Geopackage(file)
        return cls(gdf, db)

    def select_location_info(self) -> pd.DataFrame:  # pragma: no cover
        """
        Select the location information from the BRO CPT geopackage.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the location information.

        """
        raise NotImplementedError("This method is not implemented yet.")
        gcs = CptTables.geotechnical_cpt_survey
        dl = CptTables.delivered_location
        bl = CptTables.bro_location
        bp = CptTables.bro_point

        query = f"""
            SELECT
                GCS.fid,
                DL.horizontal_positioning_date, DL.horizontal_positioning_method,
                BL.crs,
                BP.x_or_lon, BP.y_or_lat
            FROM {gcs} GCS
            JOIN {dl} DL ON GCS.fid = DL.{gcs}_fk
            JOIN {bl} BL ON DL.{dl}_pk = BL.{dl}_fk
            JOIN {bp} BP ON BL.{bl}_pk = BP.{bl}_fk
            LIMIT 5
        """
        with self.db:
            self.db.query(query)
        pass



import pyproj
import pytest
from numpy.testing import assert_almost_equal

from geost import projections


@pytest.mark.unittest
def test_get_horizontal_transformer():
    t = projections.horizontal_reference_transformer(28992, 4326)
    assert isinstance(t, pyproj.transformer.Transformer)


@pytest.mark.unittest
def test_get_vertical_transformer():
    t = projections.vertical_reference_transformer(28992, 5709, 5710)
    assert isinstance(t, pyproj.transformer.Transformer)


@pytest.mark.unittest
def test_projected_to_geographic():
    t = projections.horizontal_reference_transformer(28992, 4326)
    x, y = 140_000, 440_000
    lon, lat = t.transform(x, y)
    assert_almost_equal(lon, 5.169024734261481)
    assert_almost_equal(lat, 51.948244431668776)


@pytest.mark.unittest
def test_geographic_to_projected():
    t = projections.horizontal_reference_transformer(4326, 28992)
    lon, lat = 5.1, 51.9
    x, y = t.transform(lon, lat)
    assert_almost_equal(x, 135233.28064173012)
    assert_almost_equal(y, 434649.01993844495)
